--
-- Copyright (C) 2016 HBM, Inc. All rights reserved.
--

local cjson = require 'cjson'
local log   = require 'log'

local calibration = {}

--
-- Examines the specified array of calibration steps and returns the
-- first step with the given purpose. The 'steps' argument is the
-- 'calibration' array from the setup stanza.
--
function calibration.find_cal_step(steps, purpose)
    if not steps or type(steps) ~= 'table' then return end
    for _, step in ipairs(stanza.calibration) do
        if step.purpose == purpose then
            return step
        end
    end
end

--
-- Calculates a cumulative scale and offset by sequentially applying each
-- calibration step in the specified array of steps. The 'steps' argument
-- is the 'calibration' array from the setup stanza.
--
function calibration.flatten_cal_steps(steps)
    local scale = 1
    local offset = 0
    if type(steps) == 'table' then
        for _, step in ipairs(steps) do
            local next_scale, next_offset =
                calibration.get_scale_and_offset(step)
            offset = next_scale * offset + next_offset
            scale = next_scale * scale
        end
    end
    return scale, offset
end

--
-- Gets the effective scale and offset value from a calibration step. The
-- 'step' argument is a single entry in the 'calibration' array of a setup
-- stanza.
--
function calibration.get_scale_and_offset(step)
    local scale = 1
    local offset = 0
    if step.mode == 'defined_two_point' or
            step.mode == 'experimental_two_point' then
        local e = step.electrical
        local p = step.physical
        scale = (p[2] - p[1]) / (e[2] - e[1])
        offset = p[1] - scale * e[1]
    elseif step.mode == 'defined_zero_span' or step.mode == 'shunt' then
        scale = step.physical_span / step.electrical_span
        offset = -scale * step.electrical_zero
    elseif step.mode == 'defined_slope_intercept' or step.mode == nil then
        scale = step.slope or 1
        offset = step.intercept or 0
    elseif step.mode ~= 'strain_gage' and step.mode ~= 'internal' then
        log:log(log.ERR, string.format(
            'Unknown calibration mode: %s calibration=%s',
            step.mode, cjson.encode(step)))
    elseif step.mode == 'defined_sensitivity_factor' then
        scale = step.sensitivity or 1
        offset = 0
    end
    return scale, offset
end

--
-- Returns an array of calibration steps containing only those steps for
-- which 'applied' is NOT set to true. The 'steps' argument is the
-- 'calibration' array from the setup stanza.
--
function calibration.get_unapplied_steps(steps)
    local result = {}
    if type(steps) == 'table' then
        for _, step in ipairs(steps) do
            if not step.applied then
                result[#result + 1] = step
            end
        end
    end
    return result
end

return calibration
