
local __ = require '__'
local log = require 'log'
log:open('dio_test.lua', log.USER)
local ev = require 'ev'
local loop = ev.Loop.default
local u = require 'unix'
local timestamps = require 'timestamps'
local dio = require 'dio'
local ffi = require 'ffi'
local listener = require 'listener'
local dbus = require 'dbus'
local sample_clock = require 'sample_clock'
local cca_handler = require 'cca_handler'
local mmio = require 'armada_mmio'
local bus = dbus.bus_get('system')
local timing = assert(bus:auto_proxy('com.somat.armada.Timing',
                                     '/com/somat/armada/Timing'))

local clkfreq = sample_clock.master_rate

local cmd_timeout = 0.01

local on = true
local off = false

local function sleep(t)
    ev.Timer.new(function() loop:unloop() end, t, 0):start(loop)
    loop:loop()
end

local function cable_detect(devname)
    local dio_channel = {}
    if devname == '/dev/cca8' then
        dio_channel = dio:new(devname, "dio-qc-out.bin", clkfreq)
    else
        dio_channel = dio:new(devname, "dio.bin", clkfreq)
    end
    dio_channel:get_cable_detect_state()
    local msg = assert(listener({dio_channel}, cmd_timeout))
    local plug = {[0]="UNPLUGGED", [1]="PLUGGED"}
    print("DIO cable detect with " .. devname .. " is ",
          plug[msg.channel_specific[0]])
end

local function display_event_messages(channel_obj, name, element)
    if element.events.cable_unplugged == 1 then
        print("DIO CBL_UNPLUGGED")
    end
    if element.events.cable_plugged == 1 then
        print("DIO CBL_PLUGGED")
    end
    if element.events.index_mark_fall == 1 then
        print("DIO index_mark_fall")
    end
    if element.events.sync_input0_fall == 1 then
        print("DIO sync_input0_fall")
    end
    if element.events.sync_input1_fall == 1 then
        print("DIO sync_input1_fall")
    end
    if element.events.sync_input2_fall == 1 then
        print("DIO sync_input2_fall")
    end
    if element.events.index_mark_rise == 1 then
        print("DIO index_mark_rise")
    end
    if element.events.sync_input0_rise == 1 then
        print("DIO sync_input0_rise")
    end
    if element.events.sync_input1_rise == 1 then
        print("DIO sync_input1_rise")
    end
    if element.events.sync_input2_rise == 1 then
        print("DIO sync_input2_rise")
    end
    if element.events.match_closed0 == 1 then
        print("DIO output 1 is off")
    end
    if element.events.match_open0 == 1 then
        print("DIO output 1 is on")
    end
    if element.events.force_closed0 == 1 then
        print("DIO output 1 is forced off")
    end
    if element.events.force_open0 == 1 then
        print("DIO output 1 is forced on")
    end
    if element.events.match_closed1 == 1 then
        print("DIO output 2 is off")
    end
    if element.events.match_open1 == 1 then
        print("DIO output 2 is on")
    end
    if element.events.force_closed1 == 1 then
        print("DIO output 2 is forced off")
    end
    if element.events.force_open1 == 1 then
        print("DIO output 2 is forced on")
    end
end

local function dio_events_test()
    local dio_channel = dio:new('/dev/cca8', "dio-qc-out.bin", clkfreq)
    dio_channel.element_action_func = display_event_messages
    -- 5V on
    io.write("Turning on internal 5V Power\n")
    dio_channel:txdr_power_on_5v()
    assert(listener({dio_channel}, cmd_timeout))

    -- start test with both DIO out ports closed
    dio_channel:set_output(1, off)
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:set_output(2, off)
    assert(listener({dio_channel}, cmd_timeout))

    dio_channel:event_setup({
                      cable_plugged = true,
                      cable_unplugged = true,
                      output1_off = true,
                      output1_on = true,
                      output1_forced_off = true,
                      output1_forced_on = true,
                      output2_off = true,
                      output2_on = true,
                      output2_forced_off = true,
                      output2_forced_on = true,
                      })
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:get_current_ts()
    local msg = assert(listener({dio_channel}, cmd_timeout))
    local start_time = msg.ts + (clkfreq / 10)
    dio_channel:event_start(start_time, 10, 10, 128)
    assert(listener({dio_channel}, cmd_timeout))

    local outputs_on = false
    for i = 1, 10 do
        if outputs_on then
            io.write("Set Ouputs OFF\n")
            dio_channel:set_output(1, off)
            assert(listener({dio_channel}, cmd_timeout))
            dio_channel:set_output(2, off)
            assert(listener({dio_channel}, cmd_timeout))
            outputs_on = false
        else
            io.write("Set Ouputs ON\n")
            dio_channel:set_output(1, on)
            assert(listener({dio_channel}, cmd_timeout))
            dio_channel:set_output(2, on)
            assert(listener({dio_channel}, cmd_timeout))
            outputs_on = true
        end
        sleep(1)
    end

    dio_channel:event_stop()
    assert(listener({dio_channel}, cmd_timeout))

    -- 5V off
    io.write("5V Power OFF\n")
    dio_channel:txdr_power_off()
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:halt()
    print("Done")
    os.exit(0)
end

-- The following test does:
--     closes the specified DIO out
--     turns on the 5V power source at user's command
--     opens the specified DIO out at user's command
--     closes the specified DIO out at user's command
--     turns off the 5V power source at user's command
local function interactive_dio_out_test(devname)
    local dio_channel = dio:new(devname, "dio-qc-out.bin", clkfreq)
    io.write("Select DIO output number - 1 or 2:  ")
    local output_nbr = io.read("*n")
    if not output_nbr then
        print(string.format("Must input 1 or 2, not %s", output_nbr))
        os.exit(0)
    end
    -- start test with both DIO out ports closed
    dio_channel:set_output(1, false)
    dio_channel:set_output(2, false)
    assert(listener({dio_channel}, cmd_timeout))
    -- 5V on
    io.write("Turning on internal 5V power\n") io.read()
    dio_channel:txdr_power_on_5v()
    assert(listener({dio_channel}, cmd_timeout))
    -- open DIO out
    io.write("Ready to turn ON the DO[" .. output_nbr .. "]") io.read()
    dio_channel:set_output(output_nbr, true)
    assert(listener({dio_channel}, cmd_timeout))
    -- close DIO out
    io.write("Ready to turn OFF the DO[" .. output_nbr .. "]") io.read()
    dio_channel:set_output(output_nbr, false)
    assert(listener({dio_channel}, cmd_timeout))
    -- 5V off
    io.write("Ready to turn OFF TXDR Power") io.read()
    dio_channel:txdr_power_off()
    assert(listener({dio_channel}, cmd_timeout))
    io.write("Off now, ready to halt dio channel") io.read()
    dio_channel:halt()
    print("Done")
    os.exit(0)
end

local function interactive_power_test(devname)
    local dio_channel = dio:new(devname, "dio-qc-out.bin", clkfreq)
    io.write("Ready to turn on 5V") io.read()
    dio_channel:txdr_power_on_5v()
    assert(listener({dio_channel}, cmd_timeout))
    io.write("Ready to boost to 12V") io.read()
    dio_channel:txdr_power_on_12v()
    assert(listener({dio_channel}, cmd_timeout))
    io.write("Ready to turn off") io.read()
    dio_channel:txdr_power_off()
    assert(listener({dio_channel}, cmd_timeout))
    io.write("Off now") io.read()
    dio_channel:halt()
    print("Done")
    os.exit(0)
end

local function get_ts_test(devname)
    local dio_channel = dio:new(devname, "dio.bin", clkfreq)
    for i=1,5 do
        io.write("Press key to get current timestamp") io.read()
        dio_channel:get_current_ts()
        local msg = assert(listener({dio_channel}, cmd_timeout))
        print(msg.ts)
    end
    dio_channel:halt()
    print("Done")
    os.exit(0)
end

local function get_sync_test(devname)
    local dio_channel = dio:new(devname, "dio.bin", clkfreq)
    for i=1,5 do
        io.write("Press key to get latched timestamp") io.read()
        dio_channel:get_sync_ts()
        local msg = assert(listener({dio_channel}, cmd_timeout))
        print(msg.ts)
    end
    dio_channel:halt()
    print("Done")
    os.exit(0)
end

local function synchronization_test(devname)
    local dio_channel = dio:new(devname, "dio.bin", clkfreq)
    dio_channel:setup_sync_ts()
    assert(listener({dio_channel}, cmd_timeout))
    local sync_time = timing:SendSyncPulse()
    dio_channel:get_sync_ts()
    local msg = assert(listener({dio_channel}, cmd_timeout))
    local offset = sync_time - msg.ts + 1
    dio_channel:ts_offset(offset)
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:setup_sync_ts()
    assert(listener({dio_channel}, cmd_timeout))
    sync_time = timing:SendSyncPulse()
    dio_channel:get_sync_ts()
    msg = assert(listener({dio_channel}, cmd_timeout))
    offset = sync_time - msg.ts
    if offset == 0 then
        print("Passed")
    else
        print("Failed, offset is", offset)
    end
    dio_channel:halt()
    print("Done")
    os.exit(0)
end

local function synchronization_loop_test(devname, count)
    local loop_count = count;
    if not loop_count then
        loop_count = 5
    else
        loop_count = tonumber(count)
    end
    local dio_channel = dio:new(devname, "dio.bin", clkfreq)
    local nbr_passed = 0
    local nbr_failed = 0

    for i = 1, loop_count do
        dio_channel:setup_sync_ts()
        assert(listener({dio_channel}, cmd_timeout))
        local sync_time = timing:SendSyncPulse()
        dio_channel:get_sync_ts()
        local msg = assert(listener({dio_channel}, cmd_timeout))
        local offset = sync_time - msg.ts + 1
        dio_channel:ts_offset(offset)
        assert(listener({dio_channel}, cmd_timeout))
        dio_channel:setup_sync_ts()
        assert(listener({dio_channel}, cmd_timeout))
        sync_time = timing:SendSyncPulse()
        dio_channel:get_sync_ts()
        msg = assert(listener({dio_channel}, cmd_timeout))
        offset = sync_time - msg.ts
        if offset == 0 then
            nbr_passed = nbr_passed + 1
--            print("Passed")
        else
            nbr_failed = nbr_failed + 1
--            print("Failed, offset is", offset)
        end
    end

    dio_channel:halt()
    print("Done:  passed = " .. nbr_passed ..
          ", failed = " .. nbr_failed)
    os.exit(0)
end

local function synchronization_all_test(devname, count)
    local loop_count = count;
    if loop_count then
        loop_count = tonumber(count)
    else
        loop_count = 1
    end
    if devname ~= "all" then
        print("Invalid parameter, <devname> must be 'all'")
        return
    end

    local dio_channel_1 = dio:new("/dev/cca5", "dio.bin", clkfreq)
    local dio_channel_2 = dio:new("/dev/cca6", "dio.bin", clkfreq)
    local dio_channel_3 = dio:new("/dev/cca7", "dio.bin", clkfreq)

    local nbr_passed = 0
    local nbr_failed = 0

    for i = 1, loop_count do
        dio_channel_1:setup_sync_ts()
        assert(listener({dio_channel_1}, cmd_timeout))
        dio_channel_2:setup_sync_ts()
        assert(listener({dio_channel_2}, cmd_timeout))
        dio_channel_3:setup_sync_ts()
        assert(listener({dio_channel_3}, cmd_timeout))

        local sync_time = timing:SendSyncPulse()

        dio_channel_1:get_sync_ts()
        dio_channel_2:get_sync_ts()
        dio_channel_3:get_sync_ts()

        local msg_1 = assert(listener({dio_channel_1}, cmd_timeout))
        local offset_1 = sync_time - msg_1.ts + 1
        local msg_2 = assert(listener({dio_channel_2}, cmd_timeout))
        local offset_2 = sync_time - msg_2.ts + 1
        local msg_3 = assert(listener({dio_channel_3}, cmd_timeout))
        local offset_3 = sync_time - msg_3.ts + 1

        dio_channel_1:ts_offset(offset_1)
        dio_channel_2:ts_offset(offset_2)
        dio_channel_3:ts_offset(offset_3)
        assert(listener({dio_channel_1}, cmd_timeout))
        assert(listener({dio_channel_2}, cmd_timeout))
        assert(listener({dio_channel_3}, cmd_timeout))

        dio_channel_1:setup_sync_ts()
        dio_channel_2:setup_sync_ts()
        dio_channel_3:setup_sync_ts()
        assert(listener({dio_channel_1}, cmd_timeout))
        assert(listener({dio_channel_2}, cmd_timeout))
        assert(listener({dio_channel_3}, cmd_timeout))

        sync_time = timing:SendSyncPulse()

        dio_channel_1:get_sync_ts()
        dio_channel_2:get_sync_ts()
        dio_channel_3:get_sync_ts()
        msg_1 = assert(listener({dio_channel_1}, cmd_timeout))
        msg_2 = assert(listener({dio_channel_2}, cmd_timeout))
        msg_3 = assert(listener({dio_channel_3}, cmd_timeout))

        offset_1 = sync_time - msg_1.ts
        offset_2 = sync_time - msg_2.ts
        offset_3 = sync_time - msg_3.ts

        if (offset_1 == 0 and
            offset_2 == 0 and
            offset_3 == 0) then
            nbr_passed = nbr_passed + 1
        else
            nbr_failed = nbr_failed + 1
        end
    end

    dio_channel_1:halt()
    dio_channel_2:halt()
    dio_channel_3:halt()
    print("Done:  passed = " .. nbr_passed ..
          ", failed = " .. nbr_failed)
    os.exit(0)
end

local function display_heartbeat_messages(channel_obj, name, ts)
    io.write("HB ") io.flush()
    channel_obj.heartbeat_count = channel_obj.heartbeat_count + 1
end

local function heartbeat_test(devname)
    local dio_channel = dio:new(devname, "dio-qc-out.bin", clkfreq)
    dio_channel.heartbeat_func = display_heartbeat_messages
    dio_channel.heartbeat_count = 0
    dio_channel:get_current_ts()
    local msg = assert(listener({dio_channel}, cmd_timeout))
    local start_time = msg.ts + clkfreq
    dio_channel:event_start(start_time - 3, 4, 16, 128)
    assert(listener({dio_channel}, cmd_timeout))
    listener({dio_channel}, 10.0)
    print("\nCleaning up")
    dio_channel:event_stop()
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:halt()
    print("Heartbeats detected = "..dio_channel.heartbeat_count)
    print("Done")
    os.exit(0)
end

local function set_thresholds(devname, threshold_hi, threshold_lo)
    local dio_channel = dio:new(devname, "dio.bin", clkfreq)
    threshold_hi = tonumber(threshold_hi)
    threshold_lo = tonumber(threshold_lo)
    dio_channel:hc_set_thresholds(threshold_hi, threshold_lo)
    assert(listener({dio_channel}, cmd_timeout))
    io.write("Thresholds set, Press key to finish") io.read()
    dio_channel:halt()
    print("Done")
    os.exit(0)
end

local function threshold_test(devname)
    local dio_channel = dio:new(devname, "dio.bin", clkfreq)
    for loops = 1,100 do
        for th = -5.0,5.0,0.01 do
            dio_channel:hc_set_thresholds(th, th * -1.0)
            assert(listener({dio_channel}, cmd_timeout))
        end
        for delay = 1,1000 do
            dio_channel:hc_set_thresholds(0, 0)
            assert(listener({dio_channel}, cmd_timeout))
        end
    end
    dio_channel:halt()
    print("Done")
    os.exit(0)
end

local function collect_adc_data_test(devname, sample_rate,
                                     samples_to_collect,
                                     threshold_hi, threshold_lo)
    local dio_channel = dio:new(devname, "dio.bin", clkfreq)
    samples_to_collect = tonumber(samples_to_collect)
    sample_rate = tonumber(sample_rate)
    threshold_hi = tonumber(threshold_hi)
    threshold_lo = tonumber(threshold_lo)
    local samples = {}
    dio_channel.element_action_func = function(channel_obj, name, pktstruct)
        samples[#samples + 1] = pktstruct.adc_data
    end
    dio_channel:get_current_ts()
    local msg = assert(listener({dio_channel}, cmd_timeout))
    dio_channel:hc_start(sample_rate, 1, msg.ts + clkfreq, 32, 128,
                         { threshold_hi = threshold_hi,
                           threshold_lo = threshold_lo })
    assert(listener({dio_channel}, cmd_timeout))

    while #samples < samples_to_collect do
        listener({dio_channel}, 0.5)
    end

    dio_channel:hc_stop()
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:halt()
    for i=1,samples_to_collect do print(samples[i]) end
    os.exit(0)
end

local function compare_all_adc_data_test(devname, sample_rate,
                                         samples_to_collect,
                                         threshold_hi, threshold_lo,
                                         amount_mismatch)
    --print("")
    --print("")
    --print("compare_all_adc_data_test parameters:  devname = " ..
    --      devname .. ", sample_rate = " .. sample_rate ..
    --      ", samples_to_collect = " .. samples_to_collect ..
    --      ", threshold_hi = " .. threshold_hi .. ", threshold_lo = " ..
    --      threshold_lo .. ", amount_mismatch = " .. tostring(amount_mismatch))

    if devname ~= "all" then
        print("Invalid parameter, <devname> must be 'all'")
        return
    end
    local amount_mismatch = tonumber(amount_mismatch) or 0
    if amount_mismatch < 0 then
        amount_mismatch = math.huge
    end
    local dio_channel_1 = {};
    local dio_channel_2 = {};
    local dio_channel_3 = {};
    cca_handler.new("/dev/cca5", dio_channel_1, dio, "dio.bin", clkfreq)
    dio_channel_1 = dio_channel_1.cca_dev
    cca_handler.new("/dev/cca6", dio_channel_2, dio, "dio.bin", clkfreq)
    dio_channel_2 = dio_channel_2.cca_dev
    cca_handler.new("/dev/cca7", dio_channel_3, dio, "dio.bin", clkfreq)
    dio_channel_3 = dio_channel_3.cca_dev

    samples_to_collect = tonumber(samples_to_collect)
    sample_rate = tonumber(sample_rate)
    threshold_hi = tonumber(threshold_hi)
    threshold_lo = tonumber(threshold_lo)

    local samples_for_1 = {}
    local samples_for_2 = {}
    local samples_for_3 = {}

    dio_channel_1.element_action_func = function(channel_obj, name, pktstruct, time_sec, time_nsec)
        local data = {}
        data.time_sec = time_sec
        data.time_nsec = time_nsec
        if pktstruct then data.adc_data = pktstruct.adc_data end
        samples_for_1[#samples_for_1 + 1] = data
    end
    dio_channel_2.element_action_func = function(channel_obj, name, pktstruct, time_sec, time_nsec)
        local data = {}
        data.time_sec = time_sec
        data.time_nsec = time_nsec
        if pktstruct then data.adc_data = pktstruct.adc_data end
        samples_for_2[#samples_for_2 + 1] = data
    end
    dio_channel_3.element_action_func = function(channel_obj, name, pktstruct, time_sec, time_nsec)
        local data = {}
        data.time_sec = time_sec
        data.time_nsec = time_nsec
        if pktstruct then data.adc_data = pktstruct.adc_data end
        samples_for_3[#samples_for_3 + 1] = data
    end

    local start_time = mmio:get_local_timestamp() + 2 * clkfreq

    dio_channel_1:hc_start(sample_rate, 1, start_time, 32, 128,
        { threshold_hi = threshold_hi,
          threshold_lo = threshold_lo })
    dio_channel_2:hc_start(sample_rate, 1, start_time, 32, 128,
        { threshold_hi = threshold_hi,
          threshold_lo = threshold_lo })
    dio_channel_3:hc_start(sample_rate, 1, start_time, 32, 128,
        { threshold_hi = threshold_hi,
          threshold_lo = threshold_lo })

    local previous_data_sec, previous_data_nsec = u.clock_gettime()
    local time_between_samples = 1 / sample_rate
    local sample_tick = 0 

    for sample_count = 1, samples_to_collect do
        assert(listener({dio_channel_1, dio_channel_2, dio_channel_3},
                        2.5))

        local current_data_sec, current_data_nsec = u.clock_gettime()
        local elapsed_tenths = (current_data_sec - previous_data_sec)
                               * 10
        elapsed_tenths = elapsed_tenths +
            (current_data_nsec - previous_data_nsec) / 100000000
        previous_data_sec = current_data_sec
        previous_data_nsec = current_data_nsec

        if elapsed_tenths > 10 then
            local sample_index = 1

            while sample_index <= #samples_for_1 and
                  sample_index <= #samples_for_2 and
                  sample_index <= #samples_for_3 do
                local adc_data1 = tonumber(samples_for_1[sample_index].adc_data)
                local adc_data2 = tonumber(samples_for_2[sample_index].adc_data)
                local adc_data3 = tonumber(samples_for_3[sample_index].adc_data)

                if (math.abs(adc_data1 - adc_data2) >= amount_mismatch) or
                   (math.abs(adc_data1 - adc_data3) >= amount_mismatch) or
                   (math.abs(adc_data2 - adc_data3) >= amount_mismatch) then
                    adc_data1 = tostring(adc_data1)
                    adc_data2 = tostring(adc_data2)
                    adc_data3 = tostring(adc_data3)
                    local pr_str = string.format("%s %s %s %s",
                        sample_tick, adc_data1, adc_data2, adc_data3)
                    print(pr_str)
                end

                sample_index = sample_index + 1
                sample_tick = sample_tick + time_between_samples
                -- Remove floating point conversion detritus
                sample_tick = tostring(sample_tick)
                if #sample_tick > 14 then
                    sample_tick = string.sub(1, 14)
                end
                sample_tick = tonumber(sample_tick)
            end
            samples_for_1 = {}
            samples_for_2 = {}
            samples_for_3 = {}
        end
    end

    dio_channel_1:hc_stop()
    dio_channel_2:hc_stop()
    dio_channel_3:hc_stop()

    assert(listener({dio_channel_1}, cmd_timeout))
    assert(listener({dio_channel_2}, cmd_timeout))
    assert(listener({dio_channel_3}, cmd_timeout))

    dio_channel_1:halt()
    dio_channel_2:halt()
    dio_channel_3:halt()

    --print("Test finished")

    os.exit(0)
end

local function collect_pin_state_test(devname, sample_rate,
                                      samples_to_collect,
                                      threshold_hi, threshold_lo)
    local dio_channel = dio:new(devname, "dio.bin", clkfreq)
    samples_to_collect = tonumber(samples_to_collect)
    sample_rate = tonumber(sample_rate)
    threshold_hi = tonumber(threshold_hi)
    threshold_lo = tonumber(threshold_lo)
    local samples = {}
    dio_channel.element_action_func = function(channel_obj, name, pktstruct)
        samples[#samples + 1] = pktstruct.pin_state
    end
    dio_channel:hc_set_thresholds(threshold_hi, threshold_lo)
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:get_current_ts()
    local msg = assert(listener({dio_channel}, cmd_timeout))
    dio_channel:ic_start(sample_rate, 1, msg.ts + clkfreq,
                         16, 128, {})
    assert(listener({dio_channel}, cmd_timeout))

    while #samples < samples_to_collect do
        listener({dio_channel}, 0.5)
    end

    dio_channel:ic_stop()
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:halt()
    for i=1,samples_to_collect do io.write(samples[i]) end
    os.exit(0)
end

local function pc_test(devname, sample_rate, samples_to_collect,
                       threshold_hi, threshold_lo)
    local dio_channel = dio:new(devname, "dio.bin", clkfreq)
    samples_to_collect = tonumber(samples_to_collect)
    sample_rate = tonumber(sample_rate)
    threshold_hi = tonumber(threshold_hi)
    threshold_lo = tonumber(threshold_lo)
    local samples = {}
    dio_channel.element_action_func = function(channel_obj, name, pktstruct)
        samples[#samples + 1] = {
            period_count_lo = pktstruct.period_count_lo,
            period_count_hi = pktstruct.period_count_hi,
            edge_count_lo = pktstruct.edge_count_lo,
            edge_count_hi = pktstruct.edge_count_hi
        }
    end
    dio_channel:hc_set_thresholds(threshold_hi, threshold_lo)
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:get_current_ts()
    local msg = assert(listener({dio_channel}, cmd_timeout))
    dio_channel:pc_start(sample_rate, 1, msg.ts + clkfreq,
                         math.floor(sample_rate / 4), 128, {})
    assert(listener({dio_channel}, cmd_timeout))

    while #samples < samples_to_collect do
        listener({dio_channel}, 0.5)
    end

    dio_channel:pc_stop()
    assert(listener({dio_channel}, cmd_timeout))
    dio_channel:halt()
    for i=1,samples_to_collect do
        local s = samples[i]
        print(s.period_count_lo, s.period_count_hi,
              s.edge_count_lo, s.edge_count_hi)
        local period_count = s.period_count_lo + s.period_count_hi
        print(s.period_count_lo / clkfreq, s.period_count_hi / clkfreq,
                period_count / clkfreq,
                s.edge_count_lo * sample_rate,
                s.edge_count_hi * sample_rate,
                s.period_count_hi / period_count,
                1 / period_count * clkfreq)
    end
    os.exit(0)
end


-- Perform simple validation of command line arguments

if #arg < 2 then
    print("Usage dio_test <device name> [ POWER_TEST|GET_TS|GET_SYNC")
    print("                              |SYNC_TEST|SYNC_TEST_ALL")
    print("                              |SYNC_LOOP_TEST")
    print("                              |HEARTBEAT_TEST|PC_TEST")
    print("                              |SET_THRESHOLDS")
    print("                              |THRESHOLD_TEST")
    print("                              |COLLECT_ADC_DATA")
    print("                              |COMPARE_ALL_ADC_DATA")
    print("                              |COLLECT_PIN_STATE")
    print("                              |EVENTS_TEST")
    print("                              |CABLE_DETECT ]")
    print("NOTE that <device name> must have the value 'all' "
          .. "for SYNC_TEST_ALL and COMPARE_ALL_ADC_DATA")
    print("(Execute \"dio_test <device name> <test name from list "
          .. "above>\" for help with test parameters)")
    os.exit(1)
end

if (arg[2] == "SET_THRESHOLDS" and #arg < 4) then
    if arg[2] == "SET_THRESHOLDS" then
        print("Usage dio_test <device name> SET_THRESHOLDS " ..
              "<threshold_hi> <threshold_lo>")
        os.exit(1)
    end
elseif (#arg < 6 and
        (arg[2] == "COLLECT_ADC_DATA" or
        arg[2] == "COMPARE_ALL_ADC_DATA" or
        arg[2] == "COLLECT_PIN_STATE" or
        arg[2] == "PC_TEST")) then
    if arg[2] == "COLLECT_ADC_DATA" then
        print("Usage dio_test <device name> COLLECT_ADC_DATA " ..
              "<sample_rate> <samples_to_collect> " ..
              "<threshold_hi> <threshold_lo>")
    elseif arg[2] == "COMPARE_ALL_ADC_DATA" then
        print("Usage dio_test <device name> COMPARE_ALL_ADC_DATA " ..
              "<sample_rate> <samples_to_collect> " ..
              "<threshold_hi> <threshold_lo> [amount mismatch allowed]")
    elseif arg[2] == "COLLECT_PIN_STATE" then
        print("Usage dio_test <device name> COLLECT_PIN_STATE " ..
              "<sample_rate> <samples_to_collect> " ..
              "<threshold_hi> <threshold_lo>")
    elseif arg[2] == "PC_TEST" then
        print("Usage dio_test <device name> PC_TEST " ..
              "<sample_rate> <samples_to_collect> " ..
              "<threshold_hi> <threshold_lo>")
    end
    os.exit(1)
end

-- Execute the requested test

if #arg > 1 then
    if arg[2] == "POWER_TEST" then
        interactive_power_test(arg[1])
    elseif arg[2] == "GET_TS" then
        get_ts_test(arg[1])
    elseif arg[2] == "GET_SYNC" then
        get_sync_test(arg[1])
    elseif arg[2] == "SYNC_TEST" then
        synchronization_test(arg[1])
    elseif arg[2] == "SYNC_LOOP_TEST" then
        synchronization_loop_test(arg[1], arg[3])
    elseif arg[2] == "SYNC_TEST_ALL" then
        synchronization_all_test(arg[1], arg[3])
    elseif arg[2] == "HEARTBEAT_TEST" then
        heartbeat_test(arg[1])
    elseif arg[2] == "SET_THRESHOLDS" then
        set_thresholds(arg[1], arg[3], arg[4])
    elseif arg[2] == "THRESHOLD_TEST" then
        threshold_test(arg[1])
    elseif arg[2] == "COLLECT_ADC_DATA" then
        collect_adc_data_test(arg[1], arg[3], arg[4], arg[5], arg[6])
    elseif arg[2] == "COMPARE_ALL_ADC_DATA" then
        compare_all_adc_data_test(arg[1], arg[3], arg[4],
                                  arg[5], arg[6], arg[7])
    elseif arg[2] == "COLLECT_PIN_STATE" then
        collect_pin_state_test(arg[1], arg[3], arg[4], arg[5], arg[6])
    elseif arg[2] == "PC_TEST" then
        pc_test(arg[1], arg[3], arg[4], arg[5], arg[6])
    elseif arg[2] == "CABLE_DETECT" then
        cable_detect(arg[1])
    elseif arg[2] == "EVENTS_TEST" then
        dio_events_test()
    else
        print("Input error:  unknown test name")
    end
end
