local ffi = require 'ffi'
local C = ffi.C

ffi.cdef[[
    struct mrb {
        char    *base;          // Base address of the ring buffer
        size_t  sz;             // Ring buffer size
        size_t  wr_offset;      // Write offset
        size_t  rd_offset;      // Read offset
        size_t  ready;          // Bytes ready to be read
    };

    struct work {
        struct mrb      *mrb;
    };

    int mkstemp(char *template);
    int unlink(const char *pathname);
    int ftruncate(int fd, unsigned long length);
    void *mmap(void *addr, size_t length, int prot, int flags, int fd, unsigned
long offset);
    int close(int);
    long sysconf(int name);

    enum {
        PROT_NONE = 0,
        PROT_READ = 1,
        PROT_WRITE = 2,
        _SC_PAGESIZE = 30,
    };

    enum {
        MAP_ANON = 0x20,
        MAP_PRIVATE = 0x2,
        MAP_FIXED = 0x10,
        MAP_SHARED = 0x1,
    };
]]

local ffi_copy = ffi.copy
local ffi_new = ffi.new
local math_min = math.min

local page_size = C.sysconf(C._SC_PAGESIZE)

local mt = {}

local mt_ctype

local function round_sz(sz)
    return sz == 0 and page_size or
        math.floor(((sz - 1) / page_size) + 1) * page_size
end

function mt:create(hint_sz)
--    local rb = ffi_new('struct mrb[1]')
    local rb = self
    local path = ffi_new('char[16]', '/tmp/rb-XXXXXX')
    local fd = C.mkstemp(path)
    if fd < 0 then
        error("Unable to create temporary file for mmap use")
    end
    C.unlink(path)

    rb.sz = round_sz(hint_sz)

    C.ftruncate(fd, rb.sz)

    rb.base = C.mmap(nil, rb.sz * 2, C.PROT_NONE, C.MAP_ANON + C.MAP_PRIVATE,
        -1, 0)
    if rb.base == nil then
        error("Unable to perform 1st step of mrb mapping")
    end
    local p = C.mmap(rb.base, rb.sz, C.PROT_READ + C.PROT_WRITE,
        C.MAP_FIXED + C.MAP_SHARED, fd, 0)
    if p ~= rb.base then
        error("Unable to perform 2nd step of mrb mapping")
    end
    local p = C.mmap(rb.base + rb.sz, rb.sz, C.PROT_READ + C.PROT_WRITE,
        C.MAP_FIXED + C.MAP_SHARED, fd, 0)
    if p ~= rb.base + rb.sz then
        error("Unable to perform 3rd step of mrb mapping")
    end

    C.close(fd)

    return rb
end

-- Function used to return the amount of data available to read
function mt:ready()
    return self.ready
end

-- Function used to return the amount of space available for writing
function mt:available()
    return self.sz - self.ready
end

-- Function used to the point from which to read data
function mt:read_addr()
    return self.base + self.rd_offset
end

-- Function used to get the pointer into which data can be written
function mt:write_addr()
    return self.base + self.wr_offset
end

-- Function used to dequeue from the buffer
function mt:read_incr(num)
    self.rd_offset = self.rd_offset + num
    self.ready = self.ready - num
    if self.rd_offset >= self.sz then
        self.rd_offset = self.rd_offset - self.sz
        self.wr_offset = self.wr_offset - self.sz
    end
end

-- Function used to adjust write pointer for buffer
function mt:write_incr(num)
    self.wr_offset = self.wr_offset + num
    self.ready = self.ready + num
end

-- Function used to read from the buffer
function mt:read(b, sz)
print("ready to read", self.ready)
    local to_be_read = math_min(self.ready, sz)
    if to_be_read > 0 then
        ffi_copy(b, self:read_addr(), to_be_read)
        self:read_incr(to_be_read)
    end
    return to_be_read
end

-- Function used to enqueue to the buffer
function mt:write(b, sz)
    local to_be_written = math_min(self:available(), sz)
    if to_be_written > 0 then
        ffi_copy(self:write_addr(), b, to_be_written)
        self:write_incr(to_be_written)
    end
    return to_be_written
end

mt.__index = mt

--mt_ctype = ffi.metatype('struct mrb', mt)
mt_ctype = ffi.metatype('struct work', mt)

local m = {}

function m:new(size)
    local buf = mt_ctype()
    return buf:create(size)
end

local t = m:new(4096)

local v = ffi.new('unsigned char[2]')

v[0] = 1
local tot = 0
for i = 0, 4999 do
    local written = t:write(v, 1)
    if written == 0 then break end
    tot = tot + 1
    if i > 10 and i < 15 then t:read(v + 1, 1) end
    v[0] = v[0] + 1
    if v[0] == 0 then v[0] = 1 end
end
print("total written:", tot)

print(string.format("%02X", t.base[4095]))
for i = 0, 10 do
print(string.format(" %02X", t.base[i]))
end

return m
