--
-- Copyright (C) 2016 HBM, Inc. All rights reserved.
--

local calibration = require 'calibration'

local mx_can = {}

-- The MX* will send us data values at least at this interval, in
-- seconds. If this time elapses without the MX* receiving a new
-- CAN message, it will send us a "no data" value.
local MIN_REPEAT = 1

-- The XML-escaped preamble statement of an XML document.
local XML_PREAMBLE_ESCAPED = '&lt;?xml version=&quot;1.0&quot;' ..
    ' encoding=&quot;UTF-8&quot;?&gt;'

local function tag(name, value)
    return '&lt;' .. name .. '&gt;' .. value .. '&lt;/' .. name .. '&gt;'
end

local function open_tag(name)
    return '&lt;' .. name .. '&gt;'
end

local function close_tag(name)
    return '&lt;/' .. name .. '&gt;'
end

local function empty_tag(name)
    return '&lt;' .. name .. ' /&gt;'
end

local function is_840(mxdev)
    return mxdev.category_name == 'mx840b'
end

function mx_can.set_configuration(mxdev, xpath, xml, on_success, on_error)

    local view_num = is_840(mxdev) and 5 or 8

    return mxdev.mx.control:set_view_data(view_num, xpath, xml,
        function(results)
            if results.result == 'success' then
                if on_success then return on_success() end
            else
                print("ERROR: " .. results.msg)
                if on_error then return on_error(results.msg) end
            end
        end)

end

function mx_can.set_canbus_settings(mxdev, connector_num,
    bit_rate, termination, power, on_success, on_error)

    local xpath = is_840(mxdev) and 'Amplifier' or 'AdditionalConnector'

    xpath = xpath .. string.format(
        '/Connector[@address=&quot;%d&quot;]',
        connector_num)
    xpath = xpath .. '/Mode[@choice=&quot;CanBus&quot;]'

    local termination_str = '0'
    if termination and termination ~= 0 then termination_str = 1 end

    local xml = XML_PREAMBLE_ESCAPED
    xml = xml .. open_tag('CanBus')
    xml = xml .. tag('Baudrate', bit_rate)
    xml = xml .. tag('Termination', termination_str)
    xml = xml .. tag('ListenOnly', '0')

    if power then
        xml = xml .. tag('Supply', tostring(power))
    end

    xml = xml .. close_tag('CanBus')

    return mx_can.set_configuration(mxdev, xpath, xml, on_success, on_error)

end

function mx_can.set_decoder_settings(mxdev,
    connector_num, decoder_num, spec, on_success, on_error)

    local xpath = is_840(mxdev) and 'Amplifier' or 'AdditionalConnector'

    xpath = xpath .. string.format(
        '/Connector[@address=&quot;%d&quot;]',
        connector_num)
    xpath = xpath .. '/Mode[@choice=&quot;CanBus&quot;]'
    xpath = xpath .. '/CanBus'
    xpath = xpath .. string.format(
        '/CanInChannel[@address=&quot;%d&quot;]',
        decoder_num)

    local msg_id = bit.band(tonumber(spec.db_msg_id), 0x1fffffff)

    local extended_frame = 0
    if msg_id >= 2048 then extended_frame = 1 end

    local bit_sequence = 0
    if spec.db_data_format:find('LSB') then bit_sequence = 1 end

    -- 0 = unsigned 32 bit
    -- 1 = signed 32 bit
    -- 2 = unsigned 64 bit
    -- 3 = signed 64 bit
    -- 4 = 32 bit real
    -- 5 = 64 bit real

    local raw_value_format = 0
    if spec.db_data_format:find('^U') then
        if spec.db_bit_length > 32 then
            raw_value_format = 2
        else
            raw_value_format = 0
        end
    elseif spec.db_data_format:find('^S') then
        if spec.db_bit_length > 32 then
            raw_value_format = 3
        else
            raw_value_format = 1
        end
    elseif spec.db_data_format:find('^F') then
        if spec.db_bit_length > 32 then
            raw_value_format = 5
        else
            raw_value_format = 4
        end
    else
        return on_error('Unrecognized database data format ' ..
            spec.db_data_format)
    end

    local scale = spec.db_scale
    local offset = spec.db_offset
    local value_min = spec.db_min
    local value_max = spec.db_max

    local user_scale, user_offset = calibration.flatten_cal_steps(spec.calibration)

    scale = scale * user_scale
    offset = user_offset + offset * user_scale
    value_min = value_min * user_scale + user_offset
    value_max = value_max * user_scale + user_offset

    local xml = XML_PREAMBLE_ESCAPED

    xml = xml .. open_tag('CanSensor')
    xml = xml .. tag('Identifier', msg_id)
    xml = xml .. tag('SignalName', spec.name)
    xml = xml .. tag('OriginOfName', '1')
    xml = xml .. tag('ExtendedFrame', extended_frame)
    xml = xml .. tag('BitSequence', bit_sequence)
    xml = xml .. tag('SignalType', '0')
    xml = xml .. tag('RawValueFormat', raw_value_format)
    xml = xml .. tag('StartBit', spec.db_bit_start)
    xml = xml .. tag('SignalLength', spec.db_bit_length);
    xml = xml .. tag('ScalingFactor', scale)
    xml = xml .. tag('ScalingOffset', offset)
    xml = xml .. tag('ModeBitSequence', '1')
    xml = xml .. tag('ModeStartbit', '0')
    xml = xml .. tag('ModeLength', '2')
    xml = xml .. tag('ModeValueFormat', '0')
    xml = xml .. tag('ModeValue', '3')
    xml = xml .. tag('PhysUnit', '100000')
    xml = xml .. empty_tag('PhysUnitUsr')
    xml = xml .. tag('Min', value_min)
    xml = xml .. tag('Max', value_max)
    xml = xml .. close_tag('CanSensor')

    return mx_can.set_configuration(mxdev, xpath, xml, on_success, on_error)

end

function mx_can.set_async_signal_settings(mxdev,
    connector_num, decoder_num, on_success, on_error)

    local xpath = is_840(mxdev) and 'Amplifier' or 'AdditionalConnector'

    xpath = xpath .. string.format(
        '/Connector[@address=&quot;%d&quot;]',
        connector_num)
    xpath = xpath .. '/Mode'
    xpath = xpath .. '/CanBus'
    xpath = xpath .. string.format(
        '/CanInChannel[@address=&quot;%d&quot;]',
        decoder_num)
    xpath = xpath .. '/AsyncSignal'

    local xml = XML_PREAMBLE_ESCAPED
    xml = xml .. open_tag('AsyncSignal')
    xml = xml .. tag('MaxRepTime', MIN_REPEAT)
    xml = xml .. tag('Decimals', '4')
    xml = xml .. tag('SignalFormat', '4')
    xml = xml .. close_tag('AsyncSignal')

    return mx_can.set_configuration(mxdev, xpath, xml, on_success, on_error)

end

function mx_can.activate_all_decoders(mxdev,
    connector_num, decoder_nums, activate, on_success, on_error)

    local xpath = is_840(mxdev) and 'Amplifier' or 'AdditionalConnector'

    xpath = xpath .. string.format(
        '/Connector[@address=&quot;%d&quot;]',
        connector_num)
    xpath = xpath .. '/Mode'

    local xml = XML_PREAMBLE_ESCAPED
    xml = xml .. open_tag('CanBus')
    for _, decoder_num in ipairs(decoder_nums) do
        xml = xml .. '&lt;CanInChannel address=&quot;' ..
            decoder_num .. '&quot;&gt;'
        xml = xml .. tag('Active', activate)
        xml = xml .. close_tag('CanInChannel')
    end
    xml = xml .. close_tag('CanBus')

    return mx_can.set_configuration(mxdev, xpath, xml, on_success, on_error)

end

return mx_can
