--
-- Copyright (C) 2015 Hottinger Baldwin Measurements, Inc.
--

local common        = require 'mx.channel_proxy.common'
local mx_can        = require 'mx.can'
local task_chain    = require 'task_chain'
local cjson         = require 'cjson'

local mxChannelProxy = class('mx840ChannelProxy', common)

function mxChannelProxy:initialize(mxDevice, channel_num, state)

    if channel_num > 0 then
        common.initialize(self, mxDevice, channel_num, state)
        return
    end

    self.mxdev              = mxDevice
    self.proxy_type         = 'mxChannelProxy'
    self.can_info           = state
    self.num_signals        = 0
    self.signal_slots       = {}
    self.queued_channels    = {}

end

function mxChannelProxy:available_channels(hostname)

    if not self.can_info then
        return common.available_channels(self, hostname)
    end

    return {
        {
            name        = 'raw_' .. hostname:gsub('%W', '_'),
            connector   = '@' .. hostname,
            description = 'CAN messages',
            data_type   = 'message_float32_ntp',
            category    = 'source/hbm_mx840b',
            sample_rate = 10
        }
    }

end

function mxChannelProxy:start_comet_stream(w, chans, sink)

    if not self.can_info then
        return common.start_comet_stream(self, w, chans, sink)
    end

end

function mxChannelProxy:start_channel_stream(w, spec, md_sink, sink)

    self.mxdev.channels_configured = self.mxdev.channels_configured or 0
    self.mxdev.queued_channels = self.mxdev.queued_channels or {}

    -- If this is a new test ID, discard any queued channels.
    if spec._test_id ~= self.mxdev.test_id then
        local virt_chans = self.mxdev.mx.control.device.virtual_channels
        for _, virt_chan in ipairs(virt_chans) do
            virt_chan.proxy.queued_channels = {}
        end
        self.mxdev.queued_channels = {}
        self.mxdev.channels_configured = 0
        self.mxdev.test_id = spec._test_id
        self.mxdev.start_callback = function () end
    end

    local function finalize()

        self.mxdev.channels_configured = self.mxdev.channels_configured + 1

        if self.mxdev.channels_configured >= spec._channel_count then
            self.mxdev.mx.stream:_establish_stream(function()
                    if #self.mxdev.queued_channels > 0 then
                        self:_prepare_test(md_sink)
                    end
                    if #self.mxdev.queued_channels < spec._channel_count then
                        local start_channel_list =
                            self.mxdev.start_channel_list[spec._test_id]
                        if start_channel_list then
                            self.mxdev:_start_channels(start_channel_list)
                        end
                    end
                end, true)
        end

    end

    if self.can_info then

        -- Queue up all channel requests until all of them have been received.
        -- We're maintaining TWO lists here: self.queued_channels, which has all
        -- channels for a given CAN connector, and self.mxdev.queued_channels,
        -- which is a global queue of all channels in the test for this MX840B
        -- device. We need the latter to enforce the 128-per-connector limit,
        -- and we need the former to know when we've received
        -- start_channel_stream() calls for all the channels so we can start up.

        local queued_channel = {
            w                   = w,
            spec                = spec,
            sink                = sink,
            md_sink             = md_sink,
            stream_reference    = string.format(
                                    'CanReceiver_Connector%d_Decoder%d.Signal1',
                                    self.can_info.channel_num,
                                    #self.queued_channels + 1)
        }

        self.queued_channels[#self.queued_channels + 1] = queued_channel
        self.mxdev.queued_channels[#self.mxdev.queued_channels + 1] =
            queued_channel

        -- When the test ends, detach all channel sinks and tell libmxstream
        -- to unsubscribe from all of the signals.
        w.cleanup = function()
            self.mxdev.active_channels = self.mxdev.active_channels or 0
            self.mxdev.active_channels = self.mxdev.active_channels - 1
            if self.mxdev.active_channels == 0 then
                local refs = {}
                local tasks = task_chain:new()
                for i, channel in ipairs(self.mxdev.queued_channels) do
                    refs[#refs + 1] = channel.stream_reference
                    self.mxdev.mx.stream:detach_sink(channel.stream_reference)
                    self.mxdev.mx.stream:inject_heartbeats(
                        channel.stream_reference, -1)
                end
                self.mxdev.mx.stream:unsubscribe_all(refs, function() end)
                self.mxdev.mx.stream:destroy_stream()
                local decoder_nums = {}
                for i = 1, 128 do decoder_nums[i] = i end
                tasks:add(function(on_success, on_error)
                    return mx_can.activate_all_decoders(self.mxdev,
                        1, decoder_nums, '0', on_success, on_error) end)
                tasks:run()
            end
        end

        finalize()

    else

        common.start_channel_stream(self, w, spec, md_sink, sink, finalize)

    end

end

function mxChannelProxy:_prepare_test(md_sink)

    -- Validate setup for each of the four CAN connectors
    local virt_chans = self.mxdev.mx.control.device.virtual_channels
    for _, virt_chan in ipairs(virt_chans) do

        -- Only support up to 128 signal slots per CAN connector
        local chans = virt_chan.proxy.queued_channels or {}
        if #chans > 128 then
            md_sink({ error_message = "Cannot support more than " ..
                "128 CAN signals on a single MX840B channel" })
            return
        end

        if #chans > 0 then
            -- All signals on a connector must have same bitrate/termination
            for _, channel in ipairs(chans) do
                if channel.spec.bit_rate ~= chans[1].spec.bit_rate then
                    md_sink({ error_message = "CAN signals on a single " ..
                        "connector must have the same bit rate" })
                    return
                end
                if channel.spec.internal_termination ~=
                    chans[1].spec.internal_termination then
                    md_sink({ error_message = "CAN signals on a single " ..
                        "connector must have the same termination setting" })
                    return
                end
                if (channel.spec.transducer_power or '0') ~=
                    (chans[1].spec.transducer_power or '0') then
                    md_sink({ error_message = "CAN signals on a single " ..
                        "connector must have the same transducer power" })
                    return
                end
            end
        end

    end

    -- Set up a chain of tasks/callbacks to perform.
    local tasks = task_chain:new()
    local xml = nil

    local virt_chans = self.mxdev.mx.control.device.virtual_channels

    -- Send XML configuration for the baud rate, termination and power
    -- supply settings for each enabled connector, one by one.
    for connector_num, virt_chan in ipairs(virt_chans) do
        for _, chan in ipairs(virt_chan.proxy.queued_channels) do
            tasks:add(function(on_success, on_error)
                return mx_can.set_canbus_settings(self.mxdev, connector_num,
                    chan.spec.bit_rate, chan.spec.internal_termination,
                    chan.spec.transducer_power or 0,
                    on_success, on_error) end)
            break
        end
    end

    -- Send XML configuration for each channel.
    local activations = {}
    for connector_num, virt_chan in ipairs(virt_chans) do
        for decoder_num, chan in ipairs(virt_chan.proxy.queued_channels) do
            tasks:add(function(on_success, on_error)
                return mx_can.set_decoder_settings(self.mxdev, connector_num,
                    decoder_num, chan.spec, on_success, on_error) end)
            tasks:add(function(on_success, on_error)
                return mx_can.set_async_signal_settings(self.mxdev,
                    connector_num, decoder_num, on_success, on_error) end)
            activations[connector_num] = activations[connector_num] or {}
            activations[connector_num][decoder_num] = decoder_num
        end
    end

    -- Set all used decoders to active.
    for connector_num, decoder_nums in pairs(activations) do
        tasks:add(function(on_success, on_error)
            return mx_can.activate_all_decoders(self.mxdev, connector_num,
                decoder_nums, '1', on_success, on_error) end)
    end

    -- The last task is to announce test readiness to the test engine
    -- for all connectors. When all channels have announced ready, the
    -- test engine will call the callback below to actually subscribe
    -- to the MX signals and begin data acquisition.
    tasks:add(
        function(on_success, on_error)
            for connector_num, virt_chan in ipairs(virt_chans) do
                for _, chan in ipairs(virt_chan.proxy.queued_channels) do
                    virt_chan.proxy:announce_ready(
                        chan.spec.connector, true, 'big')
                end
            end
        end)

    -- When all channels have announced ready, the test engine will call
    -- this function, at which time we will subscribe to the MX signals and
    -- begin data acquisition.
    self.mxdev.chain_start_callbacks = true
    local old_cb = self.mxdev.start_callback
    self.mxdev.start_callback = function(_self, start_time)
        self:_start_test(md_sink)
        if old_cb then old_cb(_self, start_time) end
    end

    -- Call the device start callback when the start time is announced
    self.start_callback = function(self, start_time)
        self.start_callback = nil
        self.mxdev.start_callback(self.mxdev, start_time)
    end

    tasks:run()

end

function mxChannelProxy:_start_test(md_sink)

    -- Set up a chain of tasks/callbacks to perform.
    local tasks = task_chain:new()

    -- Tell stream.lua/libmxstream to subscribe to all of the channels.
    tasks:add(
        function(on_success, on_error)
            local refs = {}
            for _, channel in ipairs(self.mxdev.queued_channels) do
                refs[#refs + 1] = channel.stream_reference
            end
            self.mxdev.mx.stream:subscribe_all(refs,
                function(str) on_success() end)
        end)

    -- Attach the channel data sinks (queued up with all the previous
    -- calls to start_channel_stream()) to all the subscribed MX stream
    -- channels so that libmxstream can pump incoming data to the
    -- correct test engine client for each channel.
    tasks:add(
        function(on_success, on_error)
            for _, channel in ipairs(self.mxdev.queued_channels) do
                self.mxdev.mx.stream:attach_sink(
                    channel.stream_reference,
                    channel.w.socket)
                self.mxdev.mx.stream:inject_heartbeats(
                    channel.stream_reference, 1.0)
            end
            on_success()
        end)

    -- Initialize this for mx_proxy's benefit (though we won't be using
    -- the start_channel_list at all for our purposes).
    self.mxdev.start_channel_list = self.mxdev.start_channel_list or {}

    -- Run all the tasks we just queued up.
    self.mxdev.active_channels = #self.mxdev.queued_channels
    tasks:run()

end

return mxChannelProxy
