--! \file

require 'middleclass'
local ev = require 'ev'
local ffi = require 'ffi'
local C = ffi.C
local net = require 'net_ffi'
local cjson = require 'cjson'
local config = require 'system_config'
require 'mx.structs'

local logger

local mx = class('mx')
mx.Configure = require 'mx.configure'
mx.Stream = require 'mx.stream'


-- Table of adjustments made to fractions field of NTP timestamps for the first
-- sample timestamp received in the stream.  If no entry exists, for a unit
-- 0 (zero) is used.
local ntp_ts_adjust = {
    ['0'] = {                     -- hbm_legacy
        mx1601 = 871008,
        mx1601b = 871008,
        mx1615 = 344241,
        mx1615b = 344241,
        mx1609 = -31449,
        mx1609b = -31449,
        mx411 = {low_speed = 549326, high_speed = -127990},
        mx411b = {low_speed = 549326, high_speed = -127990},
        mx460 = -761856,
        mx460b = -761856,
        mx471 = 0,
        mx471b = 0,
        mx840 = 243696,
        mx840b = 243696,
        mx878 = 0,
        mx878b = 0,
    },
    ['2'] = {                     -- decimal
        mx1601 = 871008,
        mx1601b = 871008,
        mx1615 = 344241,
        mx1615b = 344241,
        mx1609 = -31020,
        mx1609b = -31020,
        mx411 = {low_speed = 549326, high_speed = -127990},
        mx411b = {low_speed = 549326, high_speed = -127990},
        mx460 = -761856,
        mx460b = -761856,
        mx471 = 0,
        mx471b = 0,
        mx840 = 243696,
        mx840b = 243696,
        mx878 = 0,
        mx878b = 0,
    },
-- binary    ['3'] = {}
}

--[[
local level = 0
local function print_table(t, lab)
    local lab = lab or ''
    print(string.format("%s%s = {", string.rep(' ', level * 4),
        tostring(lab)))
    level = level + 1
    for k, v in pairs(t) do
        if type(v) == 'table' then
            print_table(v, k)
        else
            print(string.format("%s%s = %s", string.rep(' ', level * 4), k,
                tostring(v)))
        end
    end
    level = level - 1
    print(string.format("%s}", string.rep(' ', level * 4)))
end
]]

function mx:set_channel_sample_rate(channel_num, value, cb)
    self.control:set_channel_sample_rate(channel_num, value, cb)
end

function mx:disable_realtime_signals(channel_num)
    self.control:disable_realtime_signals(channel_num)
end

-- Function used to set the zero offsets for a group of channels
function mx:set_channels_zero_offset(channel_info, cb)
    if not channel_info or #channel_info == 0 then
        if cb then
            cb({
                result = 'success',
                msg = 'zero offsets set',
            })
        end
        self.zo_channel_info = nil
        return
    end
    self.zo_channel_info = channel_info
    local cnum = channel_info[1].connector:match("^(%d+)$")
    if not cnum then
        return
    end
    local offset = tostring(channel_info[1].offset)
    if not offset then
        return
    end
    table.remove(self.zo_channel_info, 1)
    local ret = self.control:set_channel_zero_offset(cnum, offset, function(r)
            if r and r.result and r.result == 'success' then
                self.set_channels_zero_offset(self, self.zo_channel_info, cb)
            else
                self.zo_channel_info = nil
                if cb then
                    cb(r)
                end
            end
        end)
    if not ret or not ret.result or ret.result ~= 'success' then
        local res = {
            result = 'error',
            msg = string.format("Unable to set zero offset for channel %s",
                cnum),
        }
        if cb then cb(res) end
        return res
    end
    return {
        result = 'success',
        msg = 'Setting channel zero offsets in progress',
    }
end

function mx:set_led(state, cb)
    self.control:set_led(state, cb)
end

function mx:read_teds_binary(connector, cb)
    self.control:read_teds_binary(connector, cb)
end

function mx:write_teds_binary(connector, bindata, cb)
    self.control:write_teds_binary(connector, bindata, cb)
end

function mx:get_sync_status()
    return self.synchronized
end

function mx:set_drd(state)
    return self.control:set_drd(state)
end

function mx:set_stream_error_handler(f)
    if not self.stream then return end
    if not self.old_stream_error_handler then
        self.old_stream_error_handler = self.stream.stream_error
    end
    self.stream.stream_error = f
end

function mx:restore_stream_error_handler()
    if self.stream and self.old_stream_error_handler then
        self.stream.stream_error = self.old_stream_error_handler
    end
end

-- Start monitoring stuff from the JET connection
function mx:start_jet(sync_cb, teds_cb)
    self.sync_cb = sync_cb
    self.teds_cb = teds_cb

    -- Monitor the JET stream for synchronization status events
    self.jet:fetch('status', function(path, event, data)
            if path == 'sync/status' then
                if event == 'change' or event == 'add' then
                    if self.synchronized.synced ~= nil and
                        self.synchronized.synced == data.sync then
                        return
                    end
                    logger:log(logger.INFO, cjson.encode(data or {}))
                    local map = {
                        ptp = 'PTPv2',
                        back_fw = 'Firewire',
                        none = 'none'
                    }
                    self.synchronized.synced = data.sync or false
                    self.synchronized.source = map[data.source or 'none']
                    if self.sync_cb then
                        self.sync_cb(self.synchronized)
                    end
                end
            elseif path:find("status/teds$") then
                if event == 'add' or event == 'change' then
                    if self.teds_cb then
                        local ref = path:match('fb/analogIn/(.+)/status')
                        self.teds_cb(ref, data.isValid, data.isPresent)
                    end
                end
            end
        end)
end

function mx:initialize(ip, _logger, loop, cb)
    local ok

    logger = _logger

    self.synchronized = {synced = false, source = 'none'}

    ok, self.control = pcall(mx.Configure.new, mx.Configure, ip, logger, loop,
        function()
            logger:log(logger.DEBUG,
                "Device initialization and default configuration complete")

            mx.Configure.configuration_finished_cb = nil

            -- Allow the MX model-specific file to do any final state
            -- checking and setup
            if self.control.device.read_initial_state then
                self.control.device:read_initial_state(self)
            end

            -- Control connection has been setup, setup JET connection
            self.jet = require'jet.peer'.new{ip = ip}
            if not self.jet then
                local msg =
                    string.format("Unable to establish JET connection to %s",
                    ip)
                logger:log(logger.ERR, msg)
                if cb then cb{result = 'error', msg = msg} end
                return
            end

            local module_type = self.control:get_module_type():lower()
            local drd = self.control.devinfo.current_drd
            if not ntp_ts_adjust[drd][module_type] then
                logger:log(logger.WARNING, string.format(
                    "No NTP timestamp adjustment available for %s using 0",
                    module_type))
            end

            local subtable
            if module_type:find('mx411') then
                if self.control.device.current_speed_mode and
                    self.control.device.current_speed_mode == 1 then
                    subtable = 'high_speed'
                else
                    subtable = 'low_speed'
                end
            end

            local tsa = type(ntp_ts_adjust[drd][module_type]) ~= 'table' and
                ntp_ts_adjust[drd][module_type] or
                ntp_ts_adjust[drd][module_type][subtable]

            if not self.control.device.use_libhmbstream then
                ok, self.stream =
                    pcall(mx.Stream.new, mx.Stream, ip, logger, loop, nil, tsa,
                        string.format("%s-%X", module_type,
                        tonumber(self.control.devinfo.serial_number)))
                if not ok then
                    local msg = string.format(
                        "Unable to establish data connection to %s\n%s",
                            ip, self.stream)
                    logger:log(logger.ERR, msg)
                    if cb then cb{result = 'error', msg = msg} end
                    return
                end
            end

            if not module_type:find('mx878') and
                not module_type:find('mx471') then
                local num = module_type:match('^mx(%d+)%d%d')
                for i = 1, num do
                    self.control:set_channel_teds_enable(i, config.allow_teds)
                end
            end

            if cb then cb{result = 'success', msg = 'ready'} end

        end)
    if not ok then
        local msg =
            string.format("Control connection initialization to %s failed", ip)
        logger:log(logger.ERR, msg)
        if cb then cb{result = 'error', msg = msg} end
        return
    end
    return self
end

return mx
