-- sie_merge.lua

local struct = require 'struct'
local class = require 'class'
local lxp = require 'lxp'
local config = require 'system_config'
local ffi = require 'ffi'
local C = ffi.C
local sie = ffi.load('libsie.so.0')
local net = require 'net_ffi'
local u = require 'unix'
local ev = require 'ev'
local sie_parser = require 'sie_parser'
local notify = require 'sys_notify'
local cjson = require 'cjson.safe'
local lfs = require 'lfs'
local log = require 'log'

ffi.cdef[[
    void *sie_context_new();
    void sie_release(void *);
    int sie_context_done(void *);
    void sie_release_types(void *);
    void *sie_writer_new(void *, int (*)(void *, const char *, int), void *);
    int sie_writer_write_block(void *, unsigned int, const char *, unsigned long);
    int sie_writer_xml_header(void *);
    int sie_writer_xml_string(void *, const char *, unsigned long);
    int sie_writer_flush_xml(void *);
    unsigned int sie_writer_next_id(void *, int);

    int read(int, void *, int);
    int open(const char *, int, int);
    int close(int);
    int write(int, void *, int);

    int htonl(int);
    int ntohl(int);

    char *strerror(int);
]]

local SieMerge = class()
local SieMergeInput = class()

function SieMergeInput:init()
    local function cb(f)
        return function(p, ...) f(self, ...) end
    end
    self.xml_parser = lxp.new({
                                  StartElement = cb(self.xml_start_element),
                                  EndElement = cb(self.xml_end_element),
                                  CharacterData = cb(self.xml_character_data)
                              })
    self.stack = { { } }
    self.remap = {
        test = { },
        ch = { },
        decoder = { },
        group = { }
    }
end

local htonl = C.ntohl
local ntohl = C.ntohl

function SieMergeInput:add_sie_block(block, size)
    local group = ntohl(ffi.cast('int *', block)[1])
    local payload = ffi.string(block, size):sub(13, -9)
    if group == 0 then
        self.xml_parser:parse(payload)
    else
        local remap = self.remap.group[tostring(group)]
        if remap then
            self.parent:output_block(remap, payload)
        end
    end
end

local tinsert, tremove, getn = table.insert, table.remove, table.getn

function SieMergeInput:xml_start_element(tag, attr)
    local newelement = {tag = tag, attr = attr}
    -- print('start', tag, getn(self.stack))
    tinsert(self.stack, newelement)
end

function SieMergeInput:xml_end_element(tag)
    local element = tremove(self.stack)
    assert(element.tag == tag)
    local level = getn(self.stack)
    -- print('end', tag, level)
    if level > 2 then
        tinsert(self.stack[level], element)
    else
        self:process_toplevel(element)
    end
end

function SieMergeInput:xml_character_data(text)
    local element = self.stack[getn(self.stack)]
    if element.tag ~= 'tag' then
        return
    end
    local n = getn(element)
    if type(element[n]) == "string" then
        element[n] = element[n] .. text
    else
        tinsert(element, text)
    end
end

local function wrap_path(el)
    local out = el
    if type(el) == 'table' then
        if el.attr.dim then
            out = { out; tag = 'dim', attr = { 'id'; id = el.attr.dim } }
            el.attr.dim = nil
        end
        if el.attr.ch then
            out = { out; tag = 'ch', attr = { 'id'; id = el.attr.ch } }
            el.attr.ch = nil
        end
        if el.attr.test then
            out = { out; tag = 'test', attr = { 'id'; id = el.attr.test } }
            el.attr.test = nil
        end
        local to_remove = { }
        for i,v in ipairs(el.attr) do
            if v == 'dim' or v == 'ch' or v == 'test' then
                table.insert(to_remove, i)
            end
        end
        repeat
            local i = table.remove(to_remove)
            if i then
                table.remove(el.attr, i)
            end
        until not i
        for i,v in ipairs(el) do
            el[i] = wrap_path(v)
        end
    end
    return out
end

function SieMergeInput:process_toplevel(el)
    el = wrap_path(el)
    if (el.tag == 'tag' and (el.attr.id == 'sie:xml_metadata' or
                             el.attr.id == 'sie:block_index')) or
       (el.tag == 'decoder' and (el.attr.id == '0' or el.attr.id == '1')) then
        return
    end
    self:process_element(el)
    self:remap_xml(el)

    local out = { }
    self:output_xml(el, function(text) table.insert(out, text) end)
    self.parent:output_xml(table.concat(out))
end

function SieMergeInput:process_element(el)
    if el.tag == 'tag' and el.attr.group then
        self:remap_id('group', el.attr.group)
    elseif el.tag == 'test' then
        self:remap_id('test', el.attr.id)
    elseif el.tag == 'ch' then
        self:remap_id('ch', el.attr.id)
        if el.attr.group then
            self:remap_id('group', el.attr.group)
        end
    elseif el.tag == 'dim' then
        if el.attr.group then
            self:remap_id('group', el.attr.group)
        end
    elseif el.tag == 'decoder' then
        self:remap_id('decoder', el.attr.id)
    end

    for i,v in ipairs(el) do
        if type(v) == 'table' then
            self:process_element(v)
        end
    end
end

local remappings = {
    ch = { base = 'ch', group = 'group', id = 'ch' },
    data = { decoder = 'decoder' },
    decoder = { id = 'decoder' },
    dim = { group = 'group' },
    tag = { decoder = 'decoder', group = 'group' },
    test = { id = 'test' },
    xform = { index_ch = 'ch' }
}

function SieMergeInput:remap_xml(el)
    local remap = remappings[el.tag]
    if remap then
        for i,v in pairs(el.attr) do
            local remap_as = remap[v]
            if remap_as then
                -- print('remap_xml', v, el.attr[v], '->',
                --       self.remap[remap_as][el.attr[v]])
                el.attr[v] = self.remap[remap_as][el.attr[v]]
            end
        end
    end

    for i,v in ipairs(el) do
        if type(v) == 'table' then
            self:remap_xml(v)
        end
    end
end

local function html_escape(str)
    return str ~= nil, str and str:gsub('[&"<]', {
                        ['&'] = '&amp;',
                        ['"'] = '&quot;',
                        ['<'] = '&lt;'
                    })
end

function SieMergeInput:output_xml(el, outfunc, indent)
    if type(el) == 'table' then
        outfunc('<' .. el.tag)
        for i,v in ipairs(el.attr) do
            local ok, s = html_escape(el.attr[v])
            if not ok then
                -- LuaJIT issue?  Possibly!
--[[
                log:log(log.NOTICE, string.format(
                    "v = %s, el.attr[v] = %s, el.attr = '%s'",
                    v, tostring(el.attr[v]), cjson.encode(el.attr)))
]]
            else
                outfunc(' ' .. v .. '="' .. s .. '"')
            end
        end
        if #el == 0 then
            outfunc('/>')
        else
            outfunc('>')
            for i,v in ipairs(el) do
                self:output_xml(v, outfunc)
            end
            outfunc('</' .. el.tag .. '>')
        end
    else
        local ok, s = html_escape(el)
        if ok then
            outfunc(s)
        end
    end
end

function SieMergeInput:remap_id(type, from, to)
    -- print('remap_id', type, from, to)
    if not self.remap[type][from] then
        if not to then
            to = self.parent:next_id(type)
        end
        self.remap[type][from] = tostring(to)
    end
end


local function sie_writer_write_block(writer, group, string)
    return sie.sie_writer_write_block(writer,
        ffi.new('unsigned int', tonumber(group)), string, string:len())
end
local function sie_writer_xml_string(writer, string)
    return sie.sie_writer_xml_string(writer, string, string:len())
end
local types = { group = 0, test = 1, ch = 2, decoder = 3 }
local function sie_writer_next_id(writer, type)
    return sie.sie_writer_next_id(writer, types[type])
end

local sie_ctx = sie.sie_context_new()

function SieMerge:init()
    self.total_written = 0
    if self.filename then
        self.file =
            C.open(self.filename, u.O_CREAT + u.O_RDWR + u.O_LARGEFILE, 438)
        local function outfunc(user, data, size)
            if self.write_error then return 0 end
            local amt_written = C.write(self.file, data, size)
            if amt_written ~= size then
                self.write_error = ffi.errno()
                log:log(log.WARNING, string.format("Write error %s -> %d, %s",
                    tostring(size), amt_written,
                    ffi.string(C.strerror(self.write_error))))
                return 0
            end
            self.total_written = self.total_written + amt_written
            return amt_written
        end
        self.writer = sie.sie_writer_new(sie_ctx, outfunc, nil)
    elseif self.fd then
        -- Turn off non-blocking mode for the fid so we don't have to deal with
        -- all of that nonsense.  Maybe someday we will deal with it.
        u.resetfl(self.fd, u.O_NONBLOCK)
        local function outfunc(user, data, size)
            if self.write_error then return 0 end
            local amt_written = C.write(self.fd, data, size)
            if amt_written ~= size then
                self.write_error = ffi.errno()
                log:log(log.WARNING, string.format("Write error %s -> %d, %s",
                    tostring(size), amt_written,
                    ffi.string(C.strerror(self.write_error))))
                return 0
            end
            self.total_written = self.total_written + amt_written
            return amt_written
        end
        self.writer = sie.sie_writer_new(sie_ctx, outfunc, nil)
    else
        local function outfunc(user, data, size)
            self.total_written = self.total_written + size
            return size
        end
        self.writer = sie.sie_writer_new(sie_ctx, outfunc, nil)
    end
    sie.sie_writer_xml_header(self.writer)
end

function SieMerge:output_xml(xml)
    sie_writer_xml_string(self.writer, xml)
    sie.sie_writer_flush_xml(self.writer)
end

function SieMerge:output_block(group, payload)
    -- print('output_block', group, payload:len())
    sie_writer_write_block(self.writer, group, payload)
end

function SieMerge:close()
    sie.sie_release(self.writer)
    if self.file then
        C.close(self.file)
    end
end

function SieMerge:next_id(type)
    return sie_writer_next_id(self.writer, type)
end

function SieMerge:new_input()
    return SieMergeInput:new{parent = self}
end

-- local files = {...}
-- local outfile = table.remove(files, 1)

-- local merge = SieMerge:new{filename = outfile}
-- for i,file in ipairs(files) do
--     local f = assert(io.open(file))
--     local input = merge:new_input()
--     local buf = ''
--     repeat
--         local new = f:read(8192) or ''
--         local size
--         buf = buf .. new
--         repeat
--             if buf:len() >= 20 then
--                 size = struct.unpack('>I4', buf:sub(1, 4))
--                 if buf:len() >= size then
--                     input:add_sie_block(buf:sub(1, size))
--                     buf = buf:sub(size + 1)
--                 end
--             end
--         until buf:len() < 20 or buf:len() < size
--     until new == ''
-- end
-- merge:close()

function SieMerge:_process_local_file(uuid)
    local buf_size = 1024 * 1024
    local buf
    while buf_size > 0 do
        buf = ffi.new('unsigned char[?]', buf_size)
        if buf then break end
        buf_size = buf_size / 2
    end
    if buf_size <= 0 then
        log:log(log.WARNING, "Unable to allocate local buffer memory")
        return
    end

    local f =
        C.open('/storage/' .. uuid .. '.sie', u.O_RDONLY + u.O_LARGEFILE, 0)
    if f <= 0 then
        buf = nil
        return
    end

    local input = self:new_input()
    local buf_offset = 0
    local tot_processed = 0
    local buf_idx = 0

    while true do
        local amt2read = buf_size - buf_offset
        if amt2read == 0 then
            buf_size = buf_size * 2
            local new_buf = ffi.new('unsigned char[?]', buf_size)
            ffi.copy(new_buf, buf, buf_size / 2)
            buf = new_buf
            amt2read = buf_size - buf_offset
        end
        local amt_read = C.read(f, buf + buf_offset, amt2read)
        if amt_read <= 0 then break end

        tot_processed = tot_processed + amt_read
        buf_offset = buf_offset + amt_read
        buf_idx = 0

        while true do
            local len = buf_offset - buf_idx
            if len < 20 then break end
            local size = ntohl(ffi.cast('int *', buf + buf_idx)[0])
            if len < size then break end
            input:add_sie_block(buf + buf_idx, size)
            buf_idx = buf_idx + size
        end

        buf_offset = buf_offset - buf_idx
        if buf_offset > 0 then
            ffi.copy(buf, buf + buf_idx, buf_offset)
        end
    end
    C.close(f)
    buf = nil
end

function SieMerge:process_single_node_file(name, uuid)
    -- Get file info
    local stat = lfs.attributes('/storage/' .. uuid .. '.sie')
    if not stat then
        error("Unable to get size of file: " .. name)
        return
    end

    if not self.fd and not self.file then
        -- Only size wanted return that
        return stat.size
    end

    local fd = self.file or self.fd

    -- Open the one source SIE file
    local fh = io.open('/storage/' .. uuid .. '.sie', 'r')
    if not fh then
        C.close(fd)
        error("Unable to open source SIE file for: " .. name)
    end

    -- Read the entire source SIE file and write it to the output FID
    local total_written = 0
    while true do
        local got = fh:read(65536)
        if not got then break end
        local amt_written = C.write(fd, got, #got)
        if amt_written ~= #got then
            C.close(fd)
            fh:close()
            error("Write error while merging " .. name)
        end
        total_written = total_written + amt_written
    end
    C.close(fd)
    fh:close()

    -- Do a final sanity check on whether or not the amount written matches
    -- the original file size.
    if total_written ~= stat.size then
        error("Size mismatch writing merged SIE file")
    end

    return total_written
end

function SieMerge:merge_by_name(name)
    local myloop = self.in_fork and ev.Loop.new() or ev.Loop.default
    name = string.gsub(name, '.sie', '')
    local netdb = require('network_db'):new()
    local hostname = netdb:get_module_hostname()

    -- First find the UUID's of all of the SIE files in the "network" that have
    -- the same name
    local rrs = netdb:query(
        [[SELECT DISTINCT owner, rdata FROM unified WHERE type = 'TXT' AND
        (rdata LIKE 'hash: %' OR rdata LIKE 'size: %') AND
        owner IN (SELECT owner FROM unified
        WHERE owner LIKE '%._armada-files' AND type = 'TXT' AND
        rdata = 'name: ]] .. name .. ".sie');")
    local uuid
    local nodes = {}
    local estimated_size
    if not rrs or #rrs < 1 then
        error("No SIE slices found for " .. name)
        return
    end
    for _, rr in ipairs(rrs) do
        local _uuid = rr.rdata:match('^hash: (.+)$')
        if _uuid then
            nodes[#nodes + 1] = rr.owner:match('^f%d+%.(.-)%._armada%-files$')
            if uuid and _uuid ~= uuid then
                -- If there is more than 1 UUID then one or more SIE files
                -- doesn't belong with the SIE file set
                self:close()
                if self.filename and not self.filename:find('^/dev/null$') then
                    os.remove(self.filename)
                end
                error("More than one SIE file named '" .. name ..
                    "' in network")
                return
            end
            uuid = _uuid
        end
        local size = rr.rdata:match('^size: (%d+)$')
        if size then
            estimated_size = estimated_size or 0
            estimated_size = estimated_size + (tonumber(size) or 0)
        end
        -- If this node is me, remove it because the SIE file portion I have is
        -- available locally
        if nodes[#nodes] == hostname then nodes[#nodes] = nil end
    end
    if not uuid then
        error("UUD for SIE file " .. name .. " not found")
        return
    end

    if #nodes == 0 then
        -- This is either a mode 1 networked test SIE file or a mode 2
        -- networked test SIE and this is the only node SIE file
        return self:process_single_node_file(name, uuid)
    end

    -- Adjust the estimated size to allow for a block header if the merged SIE
    -- file needs to be padded because it is very large and it's actual merged
    -- size is not known yet.
    if estimated_size and estimated_size > 0 then
        estimated_size = estimated_size + 20
    end

    -- Determine whether or not the SIE file is actively being written.  If so
    -- none of the final sizing modifications can be made to the .meta file
    local sie_active
    local f = io.popen('ls -l /proc/`pidof run-test2`/fd 2>/dev/null')
    if f then
        for line in f:lines() do
            if line:find(uuid, 1, true) then
                sie_active = true
            end
        end
        f:close()
    end

    -- Get the list of nodes from the SIE metadta.  If All of the nodes
    -- are NOT online, send notification but allow the download to happen.
    local sp = sie_parser:new()
    local test_nodes = sp:get_nodes('/storage/' .. uuid .. '.sie')
    local all_slices_available
    if test_nodes then
        -- Note: nodes includes all nodes EXCEPT ME!
        if (#test_nodes - 1) ~= #nodes then
            local sn = notify:new('Incomplete SIE File')
            sn.local_only = true
            sn:notify(string.format(
                "The SIE file being downloaded is an SIE file generated " ..
                "via a test that was run in 'distributed' test mode.  The " ..
                "complete SIE file consists of %d SIE 'slices' " ..
                "distributed on that same number of XR nodes.  Only %d of " ..
                "the 'slices' are available at this time.  The remaining " ..
                "'slices' are missing because either the nodes they reside " ..
                "on are no longer connected to the network or one or more " ..
                "'slices' have been deleted from their host XR nodes.  The " ..
                "SIE file will be downloaded and will be usable but will " ..
                "not be complete.", #test_nodes, #nodes + 1))
        else
            all_slices_available = true
        end
    end

    -- If this is being called to get the final size, the SIE file is NOT
    -- actively being written, AND all of the slices are available, return the
    -- final size from the .meta file if it is available.
    if not self.fd and not self.file and not sie_active and
        all_slices_available then
        -- Check to see if the size is available in the local copy of the .meta
        -- file for the SIE file.
        local f = io.open('/storage/' .. uuid .. '.meta', 'r')
        if f then
            local cj = cjson.decode(f:read('*a'))
            f:close()
            if cj and cj.merged_size then return cj.merged_size end
        end
        if not self.fd and not self.file and estimated_size >= 50000000 then
            return estimated_size
        end
    end

    local es_sn
    if not self.fd and not self.file and estimated_size and
        estimated_size < 50000000 then

        -- Actual size will be used
        estimated_size = -1

        if estimated_size > 125000000 then
            -- Send a notification about size determination taking
            -- a long time should be shown
            es_sn = notify:new('Determining Final SIE File Size')
            es_sn.local_only = true
            es_sn.dont_close = true
            es_sn:notify(string.format(
                "The size of the SIE file is quite large and will require " ..
                "some time to determine.  Please be patient.  The download " ..
                "will start shortly."), false, true)
        end
    end

    -- Move the "master" node (which is at the end of the test nodes list)
    -- from it's position in the nodes list to the beginning of the nodes
    -- list.  This makes sure the first stuff written to the new file is the
    -- stuff from the "master" slice.
    local master_found
    for i, node in ipairs(nodes) do
        if node == test_nodes[#test_nodes] then
            local entry = table.remove(nodes, i)
            table.insert(nodes, 1, entry)
            master_found = true
        end
    end

    -- Add this node to the nodes list
    test_nodes = test_nodes or {}
    if test_nodes[#test_nodes] ~= hostname then
        nodes[#nodes + 1] = hostname
    elseif not master_found then
        table.insert(nodes, 1, hostname)
    end
            
    -- Now get and combine all of the SIE slices into a single file
    local fini

    local b_size = 1024 * 1024
    local b
    while b_size > 0 do
        b = ffi.new('unsigned char[?]', b_size)
        if b then break end
        b_size = b_size / 2
    end
    if b_size <= 0 then
        log:log(log.WARNING, "Unable to allocate remote buffer memory")
        return 0
    end

    for node_num, node in ipairs(nodes) do
        if node == hostname then
            -- Slice is locally available, get it from local file
            self:_process_local_file(uuid)
        else
            -- Slice is on remote, get it from the remote machine
            fini = nil
            local ip = netdb:get_zcip_address(node)
            if not ip then
                fini = string.format(
                    "Unable to get zcip address for node %s'", node)
                break
            end
            local s = net.socket{host = ip, port = 80}:connect()
            if not s then
                fini =
                    string.format("Unable to connect to node %s', node", node)
                break
            end
            local sbuf = net.StreamBuffer:new(s.sock, 1024*1024)
            local buf = ''
            local have_headers

            local input = self:new_input()

            local clen

            local b_idx = 0
            local b_offset = 0

            function sbuf:read_cb(ev_loop)
                local avail, new = self:read()
                if new and avail > 0 then
                    if avail + b_offset > b_size then
                        while avail + b_offset > b_size do
                            b_size = b_size * 2
                        end
                        local new_b = ffi.new('unsigned char[?]', b_size)
                        ffi.copy(new_b, b, b_offset)
                        b = new_b
                    end
                    ffi.copy(b + b_offset, new, avail)
                    b_offset = b_offset + avail
                    if not have_headers then
                        local buf = ffi.string(b, b_offset)
                        local s, e = buf:find('\r\n\r\n')
                        if not s or not e then return end
                        if not buf:find('200 OK') then
                            fini = string.format(
                                "Unable to retrieve SIE slice from %s", node)
                            return
                        end
                        clen = buf:match('Content%-Length: (%d+)')
                        if not clen then
                            fini = string.format(
                                "Unable to retrieve SIE slice from %s", node)
                            return
                        end
                        buf = buf:sub(e + 1)
                        have_headers = true
                        b_idx = e
                    else
                        b_idx = 0
                    end
                else
                    self:shutdown()
                    fini = true
                end

                while true do
                    local len = b_offset - b_idx
                    if len < 20 then break end
                    local size = ntohl(ffi.cast('int *', b + b_idx)[0])
                    if len < size then break end
                    input:add_sie_block(b + b_idx, size)
                    b_idx = b_idx + size
                end

                b_offset = b_offset - b_idx
                if b_offset > 0 then ffi.copy(b, b + b_idx, b_offset) end
            end
            
            sbuf:start(myloop)
            sbuf:write('GET /sie_storage/'..uuid..'.sie '..'HTTP/1.0\r\n\r\n')

            while not fini do myloop:loop('once') end
            if type(fini) == 'string' then break end
        end
        if type(fini) == 'string' then
            if self.filename and not self.filename:find('^/dev/null$') then
                os.remove(self.filename)
            end
            self:close()
            error(fini)
        end
    end
    b = nil

    -- Update the size in the .meta file if and only of this all slices are
    -- available AND this is NOT an active SIE file
    if not sie_active and all_slices_available and not self.write_error then
        local f = io.open('/storage/' .. uuid .. '.meta', 'r')
        if f then
            local cj = cjson.decode(f:read('*a'))
            f:close()
            if cj and not cj.merged_size then
                cj.merged_size = self.total_written
                f = io.open('/storage/' .. uuid .. '.meta', 'w')
                if f then
                    f:write(cjson.encode(cj))
                    f:close()
                end
                if estimated_size and estimated_size > 0 and
                    self.total_written < estimated_size then
                    -- Pad out large SIE files that were sent using the
                    -- estimated size for the content length.
                    -- This makes the browser progress thing work.
                    estimated_size = estimated_size - self.total_written
                    if estimated_size >= 20 then
                        self:output_block(-1,
                            string.rep(' ', estimated_size - 20))
                    else
                        es_sn = notify:new('Download Error')
                        es_sn.local_only = true
                        es_sn.dont_close = true
                        es_sn:notify(
                            "The size of the SIE file was not known at " ..
                            "the time the download was attempted and " ..
                            "was not downloaded successfully.  The " ..
                            "size of the file has now been computed and " ..
                            "can be downloaded properly.  Please retry " ..
                            "your download.")
                    end
                end
            end
        end
    end

    -- Close the file
    self:close()

    -- If this is just used to get the size and the notification is shown,
    -- remove it
    if not self.fd and not self.file and es_sn then
        es_sn:clear_notification(true)
    end

    -- Log occurrance of short write
    if self.write_error then
        if self.write_error == u.EPIPE then
            log:log(log.WARNING, string.format(
                "File '%s' was not successfully uploaded because the remote " ..
                "end connection was closed while file was being uploaded.",
                name))
        else
            log:log(log.WARNING, string.format(
                "File '%s' was not successfully uploaded because a write " ..
                "error occurred.", name))
        end
    end

    return self.total_written
end

return SieMerge
