--[[
    SXR Preprocessor

    The function exported by this module is used to preprocess SXR files from
    either the setup database or as supplied by the test start API into the
    form needed by the test engine.
]]

local cjson = require 'cjson.safe'
local setup_spec = require 'setup_spec'
local uuid = require 'uuid'
local calibration = require 'calibration'
local rational = require 'rational'
local sample_clock = require 'sample_clock'
local ffi = require 'ffi'
local C = ffi.C
local log = require 'log'
local ts = require 'timestamps'
local tablex = require 'tablex'

local hostname

require('lj_include'){
    'stdlib',
    'sys/types',
    'sys/time',
    'unistd',
}

local function cleanup_summary(stanza)
    local remove_keys = {
        'channel_sort_direction', 'channel_sort_field',
        'collect_default_value', 'data_modes_sort_direction',
        'data_modes_sort_field', 'data_storage_format',
        'delete_old_data_files_on_test_initialization',
        'enable_analog_output_inversion',
        'live_updates', 'outputs_sort_direction',
        'outputs_sort_field', 'sample_rate_correction_factor',
        'setup_created_version', 'setup_file_version',
        'setup_updated_version', 'system_max_sie_tests',
        'use_calibration_dates_locks',
        'user_info', 'user_max_sie_tests', 'chart_state', 'charts', 'id',
        'ccp_stations'
    }

    for _, key in ipairs(remove_keys) do stanza[key] = nil end

    return stanza
end

local function cleanup_source(stanza)
    local remove_keys = {
        'chart_type',
        'decimals', 'description', 'drift_compensation',
        'teds_sensor', 'valid_controls',
        'ts_block'
    }
    for _, key in ipairs(remove_keys) do stanza[key] = nil end

    if not stanza.connector:find("mx%d+") and stanza.rt_enable then
        stanza.rt_enable = nil
    end

    if not stanza.connector:find('mx%d+') or
        (stanza.measurand and stanza.measurand == '') then
        stanza.measurand = nil
    end

    return stanza
end

local function cleanup_filter(stanza)
    -- If there are keys to remove, add them here
    local remove_keys = {
    }
    for _, key in ipairs(remove_keys) do stanza[key] = nil end

    return stanza
end

local function cleanup_sink(stanza)
    return stanza
end

-- Function used to add the elapsed test runtime channel to ALL tests
local function add_elapsed_time_channel(setup)
    setup[#setup + 1] = {
        category = 'filter/elapsed_run_time',
        data_type = 'sequential_float32',
        sample_rate = 1,
        sample_rate_legacy_uncorrected = 1,
        name = 'elapsed_test_run_time',
        is_local = true,
        id = uuid.generate(),
    }
    return setup
end

-- Function used to add the SIE file size channel to SIE tests
local function add_sie_size_channel(setup)
    if not setup[1].purpose then
        setup[#setup + 1] = {
            category = 'filter/sie_size',
            data_type = 'sequential_float32',
            sample_rate = 1,
            sample_rate_legacy_uncorrected = 1,
            name = 'sie_file_size',
            is_local = true,
            id = uuid.generate(),
        }
    end
    return setup
end

-- Function used to add raw CAN channels to the test for tests with CAN signals
-- defined on CAN connectors AND the test does not explicitly include the raw
-- CAN channel the CAN signals will be extracted from. (Even if the test does
-- include the raw channel, certain parameters from signal channels such as
-- Active Query parameters, and other global summary parameters like CCP data,
-- must be stuffed into the raw channel's parameters.)
local function add_raw_can_channels(setup)

    local onboard_connectors = {}

    for _, stanza in ipairs(setup) do
        if stanza.category and stanza.category:find('^source/.+can$') then
            local c = stanza.connector:match('.*(@can%d%.cx23r%-.+)') or
                stanza.connector:match('.*(@can%d%.edaqx.-%-.+)')
            if not c then
                error(string.format(
                    "Unable to determine CAN connector from '%s'",
                    stanza.connector))
            end
            onboard_connectors[c] = onboard_connectors[c] or {}
            if stanza.category:find('msg_can$') then
                onboard_connectors[c].raw = stanza
            else
                if stanza.ccp_station_key then
                    onboard_connectors[c].ccp_stations =
                        onboard_connectors[c].ccp_stations or {}
                    onboard_connectors[c].ccp_stations[
                        stanza.ccp_station_key] = true
                end
                if onboard_connectors[c].raw then
                    onboard_connectors[c].raw.signal_count =
                        onboard_connectors[c].raw.signal_count or 0
                    onboard_connectors[c].raw.signal_count =
                        onboard_connectors[c].raw.signal_count + 1
                end
                onboard_connectors[c][#onboard_connectors[c] + 1] = stanza
            end
        end
    end

    for cc_name, cc in pairs(onboard_connectors) do
        if not cc.raw then
            local name =
                string.format('raw_%s', cc_name:match('@(.+)'):gsub('-', '_'))
            local num = #cc
            cc.raw = {
                bit_rate = cc[1].bit_rate,
                category = cc[1].category:match('^(.+/).+$') .. 'msg_can',
                connector = cc_name,
                data_type = 'message_can',
                id = uuid.generate(),
                internal_termination = cc[1].internal_termination,
                name = name,
                signal_count = num,
            }

            -- Insert this raw CAN channel into the test just after the summary
            table.insert(setup, 2, cc.raw)
        end

        -- Add all CCP station parameters to the raw channel.
        if cc.ccp_stations then
            for id, _ in pairs(cc.ccp_stations) do
                cc.raw.ccp_stations = cc.raw.ccp_stations or {}
                cc.raw.ccp_stations[id] = setup[1].ccp_stations[id]
            end
        end

        -- Add all Active Query setting to a list on the raw channel.
        local query_msgs = {}
        local total_queries = 0
        for _, db_chan in ipairs(cc) do
            if db_chan.request_rate then
                db_chan.request_rate = rational:new(
                    db_chan.request_rate):tonumber()
            end
            if db_chan.request_rate and db_chan.request_msg and
                db_chan.request_rate > 0 then
                if query_msgs[db_chan.request_msg] then
                    if query_msgs[db_chan.request_msg].rate ~=
                        db_chan.request_rate then
                        if db_chan.request_rate >
                            query_msgs[db_chan.request_msg].rate then
                            query_msgs[db_chan.request_msg].rate =
                                db_chan.request_rate
                        end
                    end
                else
                    query_msgs[db_chan.request_msg] = {
                        rate = db_chan.request_rate
                    }
                    total_queries = total_queries + 1
                    if total_queries > 75 then
                        error("More than 75 Active Query messages")
                    end
                    cc.raw._active_queries = cc.raw._active_queries or {}
                    local aq = cc.raw._active_queries
                    aq[#aq + 1] = {
                        msg = db_chan.request_msg,
                        rate = db_chan.request_rate,
                    }
                end
            end
        end
    end

    return setup
end

-- Function used to add raw GPS channels to the test for tests with GPS signals
-- defined on GPS connectors AND the test does not explicitly incluode the raw
-- GPS channel the GPS signals will be extracted from.
local function add_raw_gps_channels(setup)
    local onboard_connectors = {}

    -- First, figure out which raw GPS connectors are being used
    for _, stanza in ipairs(setup) do
        if stanza.category and stanza.category:find('^source/') and
            stanza.connector and stanza.connector:find('@gps') and
            (not stanza.connector:find('pps_offset') and
            not stanza.connector:find('timestamp')) then
                
            local c = stanza.connector:match('.*(@gps.+)$')
            if not c then
                error(string.format(
                    "Unable to determine GPS connector from '%s'",
                    stanza.connector))
            end
            onboard_connectors[c] = onboard_connectors[c] or {}
            if stanza.category:find('msg_gps$') and
                stanza.connector:find('^@gps%.') then
                onboard_connectors[c].raw = stanza
            else
                onboard_connectors[c][#onboard_connectors[c] + 1] = stanza
            end
        end
    end

    for cc_name, cc in pairs(onboard_connectors) do
        if not cc.raw then
            local name =
                string.format('raw_%s', cc_name:match('@(.+)'):gsub('-', '_'))
            cc.raw = {
                category = cc[1].category:match('^(.+/).+$') .. 'msg_gps',
                connector = cc_name,
                data_type = 'message_raw',
                id = uuid.generate(),
                internal_termination = cc[1].internal_termination,
                name = name,
            }

            -- Insert this raw CAN channel into the test just after the summary
            table.insert(setup, 2, cc.raw)
        end
    end

    for cc_name, cc in pairs(onboard_connectors) do
        for i, stanza in ipairs(cc) do
            stanza.raw_channel = cc.raw.connector
        end
    end

    return setup
end

-- For the MX proxy and edaq_proxy, we need to tell the proxy the total
-- number of channels in the test which will be serviced by that proxy,
-- so they know when /channel_stream commands have been isused for all
-- channels (to allow them to perform any final validation or setup).
local function set_channel_counts(setup)

    local get_service_from_connector = function(stanza)
        if not stanza.category or not stanza.connector then return nil end
        if stanza.category:match('^source/hbm_edaqxr.*/brg$') or
                stanza.category:match('^source/hbm_edaqxr.*/xrbrg$') or
                stanza.category:match('^source/hbm_edaqxr.*/dio/input$') or
                stanza.category:match('^sink/hbm_edaqxr.*/dio/output$') or
                stanza.category:match('^source/hbm_edaqxr.*/dio/pulse$') or
                stanza.category:match('^source/hbm_edaqxr.*/hls$') or
                stanza.category:match('^source/hbm_edaqxr.*/ntb$') or
                stanza.category:match('^source/hbm_edaqxr.*/itb$') then
            local device = stanza.connector:match('%.(edaqxr.*%-[0-9a-fA-F]+)$')
            return 'legacy.' .. device
        elseif stanza.category:find('^.*/hbm_mx') then
            local device =
                stanza.connector:match('%.(mx[0-9]+b?%-[0-9a-fA-F]+%..+)$')
            return device;
        else
            return nil
        end
    end

    -- Figure out the number of channels inuse on each MX device
    local box_map = {}
    for _, stanza in ipairs(setup) do
        local key = get_service_from_connector(stanza)
        if key then box_map[key] = (box_map[key] or 0) + 1 end
    end

    -- Propagate the channel counts to each of the MX channel stanza's
    for i, stanza in ipairs(setup) do
        local key = get_service_from_connector(stanza)
        if key then setup[i]._channel_count = box_map[key] end
    end

    return setup
end

-- Function used to process the calibration information down into a single
-- scale and offset.
local function process_calibration(setup)
    for i, stanza in ipairs(setup) do
        if stanza.calibration then
            -- Normalize the calibration stanzas
            local scale, offset = calibration.flatten_cal_steps(
                calibration.get_unapplied_steps(stanza.calibration))
            if scale ~= 1.0 or offset ~= 0.0 then
                -- These default to 1.0 and 0.0 if not set!
                setup[i].scale = scale
                setup[i].offset = offset
            end
        end
    end

    return setup
end

-- This function sets the _timescale key for all source channels that don't
-- have it set already.
local function set_timescales(setup)
    for i, stanza in ipairs(setup) do
        local c = stanza.category
        if c and c:find('^source/') and stanza.data_type then
            if stanza.sample_rate and not stanza._timescale then
                setup[i]._timescale = 1.0 / stanza.sample_rate
            else
                -- KLUDGE: Hardcoded 1/40Mhz
                setup[i]._timescale = 1.0 / sample_clock.master_rate
            end
        end
    end
    return setup
end

-- This function is used to propagate the data type of filter channels' input
-- channel(s) to the filter.  Note, the setup MUST be sorted in "declare
-- before reference" order or this doesn't work!
local function fixup_filters(setup)
    for i, stanza in ipairs(setup) do
        local c = stanza.category
        if c and c:find('^filter/') then
            if not stanza.data_type then
                setup[i].data_type =
                    get_upstream_datatype(stanza.input_channels)
            end
            if not stanza.pipe_rate then
                setup[i].pipe_rate = setup[1].pipe_rate
            end
        end
    end
    return setup
end

-- This function processes all of the MX channels that have "real_time_signal"
-- information.
local function process_realtime_signals(setup)
    for i, stanza in ipairs(setup) do
        if stanza.category and stanza.category:find('^source') then
            if stanza.real_time_signal then
                if stanza.real_time_signal.signal_1_enabled then
                    setup[i].rt_enable = 1
                else
                    setup[i].rt_enable = 0
                end
                if stanza.real_time_signal.signal_2_enabled then
                    setup[i].signal2 = stanza.real_time_signal.signal_2_settings
                else
                    setup[i].signal2 = nil
                end
    
                -- This stuff is not needed after this point
                    setup[i].real_time_signal = nil
            end
            setup[i].rt_enable = setup[i].rt_enable or 0
        end
    end
    return setup
end

local function process_mx878_channels(setup)
    local function get_source_stanza_info(input)
        for _, s in ipairs(setup) do
            if s.category and s.category:find('^source/hbm_mx') then
                if s.name == input.source_channel_name then
                    local ref, module =
                        s.connector:match('^@connector(%d+)%.(.+)$')
                    -- FIXME:  Don't know if this is correct
                    -- for MX460 channels
                    -- FIXME: This doesn't work for CAN channels
                    -- on MX471/MX840, but at this time those are
                    -- not supported as MX878 inputs.
                    ref = "AnalogIn_Connector" .. ref ..  '.Signal' ..
                        tostring(input.source_signal)
                    return ref, module
                end
            end
        end
    end

    for i, stanza in ipairs(setup) do
        if stanza.category and stanza.category:find('^sink/hbm_mx878') and
            stanza.input and type(stanza.input) == 'table' then
            setup[i].input.reference, setup[i].input.module =
                get_source_stanza_info(stanza.input)
            setup[i].input.signal = nil
            setup[i].input.connector = nil
            setup[i].input.source_signal = nil
            setup[i].input.source_channel_name = nil
        end
    end

    return setup
end

local function cleanup_setup(setup)
    -- Now remove all of the excess cruft from all of the stanzas that isn't
    -- needed by the test engine.
    for i, stanza in ipairs(setup) do
        if not stanza.category then
            -- No category, ignore the stanza
        elseif stanza.category == 'summary' then
            setup[i] = cleanup_summary(stanza)
        elseif stanza.category:find('^sink/') then
            setup[i] = cleanup_sink(stanza)
        elseif stanza.category:find('^filter/') then
            setup[i] = cleanup_filter(stanza)
        elseif stanza.category:find('^source/') then
            setup[i] = cleanup_source(stanza)
        end
    end
    return setup
end

-- Function used to get a list of XR network nodes from the channels in a test
local function get_test_nodes(setup)
    local nodes
    for _, stanza in ipairs(setup) do
        if stanza.connector and stanza.category and
            stanza.category:find('^source/') then
            local node = stanza.connector:match('%.(cx23r%-.+)$') or
                stanza.connector:match('%.(edaqx.-%-.+)$') or
                (stanza.network_node and
                    stanza.network_node:match('^(cx23r%-.+)$') or
                    stanza.network_node:match('^(edaqx.-%-.+)$'))
            nodes = nodes or {}
            nodes[node] = true
        end
    end
    local nn = {}
    for node, _ in pairs(nodes or {}) do nn[#nn + 1] = node end
    return nn
end

-- This function is used to add the "network_nodes" key to the summary stanza
-- of mode 2 networked tests.
local function add_network_nodes_list(setup)
    for i, stanza in ipairs(setup) do
        if stanza.category == 'summary' and stanza.network_mode and
            tonumber(stanza.network_mode) == 2 then
                setup[i].network_nodes = get_test_nodes(setup)
            break
        end
    end
    return setup
end

-- Function used to add network_node to stanza's that don't have them
local function add_network_node(setup)
    -- Set the network_node field of all source channels that don't already
    -- have it set.
    for i, stanza in ipairs(setup) do
        if stanza.category and stanza.category:find('^source/') and
            not stanza.network_node and stanza.connector then
            setup[i].network_node = stanza.connector:match('.+%.(.+)$')
        elseif stanza.category and not stanza.network_node and
            stanza.connector and
            stanza.category:match('^sink/hbm_edaqxr.*/dio/output$') then
            setup[i].network_node = stanza.connector:match('.+%.(.+)$')
        end
    end

    -- Helper function for use with setting network_node in filter (computed)
    -- and datamode stanzas.
    local function get_filter_network_node(stanza, setup)
        local network_node
        local network_mode = setup[1].network_mode or 1
        if network_mode ~= 2 then return hostname end
        for i, ic in ipairs(stanza.input_channels) do
            for ii, c in ipairs(setup) do
                if c.name == ic then
                    if c.category:find('^filter/') then
                        local n = get_filter_network_node(c, setup)
                        if network_node and network_node ~= n then
                            error(string.format(
                                "Channel '%s' is using channels " ..
                                "from more than one network node", stanza.name))
                        end
                        network_node = n
                    elseif not network_node then
                        network_node = c.network_node
                    elseif network_node ~= c.network_node then
                        error(string.format(
                            "Channel '%s' is using channels " ..
                            "from more than one network node", stanza.name))
                    end
                end
            end
        end
        return network_node
    end

    -- Deal with filters (computed channels)
    for i, stanza in ipairs(setup) do
        if stanza.category and stanza.category:find('^filter/') and
            not stanza.network_node then
            setup[i].network_node = get_filter_network_node(stanza, setup)
        end
    end

    -- Deal with datamodes if this is an SIE test
    if not setup[1].purpose then
        for i, stanza in ipairs(setup) do
            if stanza.category and stanza.category:find('^sink/sie') and
                not stanza.network_node and stanza.input_channels then
                setup[i].network_node = get_filter_network_node(stanza, setup)
                if stanza.trigger_channel and
                    stanza.trigger_channel ~= '' then
                    local t_node = get_filter_network_node(
                        {input_channels={stanza.trigger_channel}}, setup)
                    if setup[1].network_mode == 2 and
                        t_node ~= setup[i].network_node then
                        error(string.format(
                            "Datamode '%s' is using a trigger channel from " ..
                            "a different network node than that of the " ..
                            "input channels.", stanza.name))
                    end
                end

                -- Add pipe_rate if it doesn't exist because some need it
                if not stanza.pipe_rate then
                    setup[i].pipe_rate = setup[1].pipe_rate
                end
            end
        end
    end

    return setup
end

-- This function is used to assign SIE buffer indices to all of the channels
-- that need SIE buffers.  I'm not sure if the order in which the channels
-- show up in InField is dictated by the SIE buffer group number or by the
-- order in which the channel XML appears in the SIE file.
local function assign_sie_buffer_offsets(setup)
    local sid = 1
    for i, stanza in ipairs(setup) do
        if stanza.category then
            if stanza.category:find('^source/') then
                setup[i].sid = sid
                -- Two stream id's are needed for each source channel.  One for
                -- the normal rate stream and one for the "low speed" rate.
                sid = sid + 2
            elseif stanza.category:find('^filter/') then
                -- Most all filter channels use only one stream id
            elseif stanza.category:find('^sink/sie') then
                -- Sinks don't use ANY stream ids!
            end
        end
    end
    return setup
end

-- Get a list of "test_order" indexes that specifies the "display" order
local function get_display_order(setup)
    local sxr = {}
    for _, s in ipairs(setup) do sxr[#sxr + 1] = s end
    table.sort(sxr, function(a, b)
            if not a.sort_order then return true
            elseif not b.sort_order then return false
            else return tonumber(a.sort_order) < tonumber(b.sort_order) end
        end)

    local display_order = {}
    for _, stanza in ipairs(setup) do
        local ok, so = pcall(tonumber, stanza.sort_order)
        if ok and so and so > 0 then
            display_order[#display_order + 1] = so
        end
    end
    return display_order, sxr
end

local function get_start_time_delay(setup)
    -- Set the start time delay for the test
    local start_time_delay = 0
    local enable_sync_stops
    if setup[1].purpose then
        for _, stanza in ipairs(setup) do
            if stanza.category and stanza.category:find('^filter/') then
                start_time_delay = setup[1].test_start_delay or 3
                enable_sync_stops = true
                break
            end
            if stanza.connector and 
                (stanza.connector:find('@can%d%..+') or
                stanza.connector:find('@gps%..+')) then
                start_time_delay = setup[1].test_start_delay or 2
                enable_sync_stops = true
                break
            end
        end
    else
        start_time_delay = setup[1].test_start_delay or 0
        enable_sync_stops = true
    end
    setup[1].test_start_delay = start_time_delay
    setup[1].enable_sync_stops = enable_sync_stops

    -- Propagate the start time delay to channels because some may use it for
    -- error checking purposes.
    for i, stanza in ipairs(setup) do
        if stanza.category and stanza.category:find("^source/") then
            if stanza.category:find("^source/hbm_edaqxr.*/digital_input") and
                stanza.data_type:find("^message_float") then
                setup[i].stale_data_expiration_time =
                    stanza.stale_data_expiration_time or 1
                setup[i].invalid_data_output_value =
                    stanza.invalid_data_output_value or -1
            end

            setup[i].start_time_delay = start_time_delay
        end
    end
    return setup
end


local function fixup_calibration(setup)
    for i, stanza in ipairs(setup) do
        if stanza.category and stanza.category:find("^source/") then

            -- For all MX source channels set "applied" in all of the
            -- calibration stanza's to true and for all other source channel
            -- types, remove "applied" from all calibration stanza's
            local lower_category = stanza.category:lower()
            if (lower_category:find('^source/hbm_mx') or
                lower_category:match('^source/hbm_edaqxr.*/brg$') or
                lower_category:match('^source/hbm_edaqxr.*/dio/pulse$') or
                lower_category:match('^source/hbm_edaqxr.*/hls$') or
                lower_category:match('^source/hbm_edaqxr.*/itb$') or
                lower_category:match('^source/hbm_edaqxr.*/ntb$')) and
                stanza.calibration then
                for j, _ in ipairs(stanza.calibration) do
                    setup[i].calibration[j].applied = true
                end
            elseif stanza.calibration then
                for j, _ in ipairs(stanza.calibration) do
                    setup[i].calibration[j].applied = nil
                end
            end
        end
    end

    return setup
end

-- This function is used to append the "network_node" field of MX channel
-- stanzas to the "connector" field.
local function fixup_mx_connectors(setup)
    for i, stanza in ipairs(setup) do
        if stanza.connector and stanza.connector:find("mx%d+") and
            stanza.network_node then
            setup[i].connector = stanza.connector .. '.' .. stanza.network_node
        end
    end
    return setup
end

-- This function is used to set the 'min_time_between_notifications' and
-- 'max_notifications_per_test_run' parameters in a setup.
local function fixup_notification_throttling(setup)
    for i, stanza in ipairs(setup) do
        if stanza.category == 'summary' then
            for key, info in pairs{
                min_time_between_notifications =
                    {default = 60, min = 30, max = 120},
                max_notifications_per_test_run =
                    {default = 10, min = 1, max = 50}} do

                setup[i][key] = stanza[key] or info.default
                setup[i][key] = rational:new(stanza[key]):tonumber()
                if setup[i][key] < info.min or setup[i][key] > info.max then
                    error(string.format(
                        "Invalid '%s' value (%d) - must be between %d and %d",
                        key, setup[i][key], info.min, info.max))
                    break
                end
            end
        end
    end
    return setup
end

local function get_network_node(stanza)
    -- This may need more than this
    return stanza.network_node
end

local function get_datamode_network_node_info(setup, stanza)
    local nni = {}

    for _, ic in ipairs(stanza.input_channels) do
        for i, s in ipairs(setup) do
            if s.name and s.name == ic then
                -- This is one of the input channels, get it's network node
                local nn = get_network_node(s)
                nni[nn] = nni[nn] or {}
                nni[nn][ic] = true
            end
        end
    end

    return nni
end

local function split_datamode(setup, stanza)
    local new_stanzas = {}

    -- Get how many network nodes are used AND the input channels used from
    -- each.
    local network_node_info = get_datamode_network_node_info(setup, stanza)

    -- Generate one stanza per network node.  Each with only the channels that
    -- are on that network node
    for nn, nn_ic in pairs(network_node_info) do
        new_stanzas = new_stanzas or {}
        new_stanzas[#new_stanzas + 1] = tablex.deepcopy(stanza)
        new_stanzas[#new_stanzas].network_node = nn

        for j, ic in ipairs(stanza.input_channels) do
            if not nn_ic[ic] then
                new_stanzas[#new_stanzas].input_channels[ic] = nil
            end
        end
    end

    -- TODO: Deal with the trigger channel, if any

    return new_stanzas
end

local function do_mode2_networked_processing(setup)
    for i, stanza in ipairs(setup) do
        if stanza.category == 'summary' and
            (not stanza.network_mode or not tonumber(stanza.network_mode) or
            tonumber(stanza.network_mode) ~= 2) then
            return setup
        end
    end

    -- If we get here this is a mode 2 networked test
    local new_setup = {}
    for _, stanza in ipairs(setup) do
        if stanza.category then
            if stanza.category:find('^filter/') then
                if stanza.input_channels and #stanza.input_channels > 1 then
                    -- TODO: Duplicate this channel on all nodes or 
                    new_setup[#new_setup + 1] = stanza
                else
                    new_setup[#new_setup + 1] = stanza
                end
            elseif stanza.category:find('^sink/sie') then
                -- TODO: Split the data mode if needed
                local new_stanzas = split_datamode(setup, stanza)
                for _, s in ipairs(new_stanzas) do
                    new_setup[#new_setup + 1] = s
                end
            else
                -- All other nodes are on the network node specified
                new_setup[#new_setup + 1] = stanza
            end
        end
    end

    return new_setup
end

-- Function used to do any preprocessing needed for legacy eDAQ layer channels
local function process_legacy_edaq_layer_channels(setup)
    for i, stanza in ipairs(setup) do
        if stanza.category and stanza.category:find('^source') and
            stanza.category:find('^source/hbm_edaqx.*/') then
            setup[i].sample_rate_domain = setup[1].sample_rate_domain
        elseif stanza.category and stanza.category:find('^sink') and
            stanza.category:find('^sink/hbm_edaqx.*/output$') then
            setup[i].sample_rate_domain = setup[1].sample_rate_domain
        end
    end

    return setup
end


-- Function used to adjust calibration for uncalibrated channels
local function process_uncalibrated_channels(setup)
    for i, stanza in ipairs(setup) do
        -- Deal with "uncalibrated" channels
        if (stanza.calibration_date and
            stanza.calibration_date:lower() == 'not_calibrated') or
            stanza.force_electrical_units_data_stream then

            -- Set scaling to unity (scale=1, offset=0)
            setup[i].calibration = {
                {
                    intercept = 0,
                    purpose = 'sensor_scaling',
                    slope = 1,
                    mode = 'defined_slope_intercept',
                    units = stanza.calibration[1].units and
                        (stanza.calibration[1].units:lower():find('degrees') and
                            'Degrees C' or stanza.calibration[1].units),
                },
                {
                    purpose = 'rezero',
                    slope = 1,
                    intercept = 0,
                    mode = 'defined_slope_intercept',
                },
            }

            -- For BRG channels, must set the measurement range (range_min
            -- and range_max) to some reasonable defaults, and those defaults
            -- depend on the sensor input mode.
            if stanza.connector:match('@x*brg_%d%.') and stanza.purpose then
                if stanza.sensor_input then
                    if stanza.sensor_input.mode == 'volt' then
                        setup[i].range_min = -10
                        setup[i].range_max = 10
                    elseif stanza.sensor_input.mode == 'bridge' then
                        setup[i].range_min = -200
                        setup[i].range_max = 200
                    elseif stanza.sensor_input.mode == 'diff_ampl' then
                        setup[i].range_min = -10000
                        setup[i].range_max = 10000
                    end
                end

            -- For HLS channels, must set the measurement range (range_min
            -- and range_max) to some reasonable defaults, and those defaults
            -- depend on the sensor input mode (and SMART modules have
            -- different defaults).
            elseif stanza.connector:match('@x*hls_%d%.') and stanza.purpose then
                if stanza.sensor_input then
                    if stanza.sensor_input.mode == 'volt' then
                        if stanza.sensor_input.min == -80 and
                            stanza.sensor_input.max == 80 then
                            setup[i].range_min = -80
                            setup[i].range_max = 80
                        else
                            setup[i].range_min = -10
                            setup[i].range_max = 10
                        end
                    elseif stanza.sensor_input.mode == 'iepe' then
                        setup[i].range_min = -10
                        setup[i].range_max = 10
                    elseif stanza.sensor_input.mode == 'bridge' then
                        setup[i].range_min = -36
                        setup[i].range_max = 36
                    elseif stanza.sensor_input.mode == 'thermocouple' then
                        setup[i].range_min = -200
                        setup[i].range_max = 1372
                    end
                end

            end

        end
    end

    return setup
end

-- Function used to propagate some of the summary fields to the other stanzas
local function propagate_summary_info(setup)
    for i, stanza in ipairs(setup) do
        if stanza.category and stanza.category ~= 'summary' then
            for _, k in ipairs{'purpose', 'duration',
                'sample_rate_domain', 'digital_display_type',
                'digital_display_setup_id', 'pipe_rate',
                'collect_data_mode_name', 'enable_analog_output_inversion'} do
                if setup[1][k] then setup[i][k] = setup[1][k] end
            end
        end
    end
    return setup
end

-- Function used to enable prerun zeroing on channels that have it enabled
local function set_prerun_zero(setup)
    for i, stanza in ipairs(setup) do
        if stanza.prerun_balance and stanza.prerun_balance.mode and
            stanza.prerun_balance.mode == 'normal_test_starts' then
            setup[i].do_prerun_zero = true
        end
    end
    return setup
end


-- Make sure any booleans that have been supplied as strings get converted to
-- real Lua booleans.
local function fixup_booleans(stanza)
    for k, v in pairs(stanza) do
        if type(v) == 'table' then
            stanza[k] = fixup_booleans(v)
        elseif type(v) == 'string' then
            if v:lower() == 'true' then
                stanza[k] = true
            elseif v:lower() == 'false' then
                stanza[k] = nil
            end
        end
    end
    return stanza
end


-- Function use to preprocess the setup such that it is has ALL of the stuff
-- done to it that the test engine needs so the test engine itself doesn't have
-- to do ANY preprocssing of the setup.  Things done here are validations,
-- sorting into test engine order, insertion of channels if needed (raw GAN,
-- GPS, elapsed time, etc.), calculation of final scale/offset, removal of all
-- of the non-test engine essential information (charts, descrptions, etc.).
return function(setup, host)
    if type(setup) == 'string' then
        -- If the argument is a string then assume it is JSON and try to
        -- decode it.
        setup = cjson.decode(setup)
        if not setup then
            error("setup preprocessor: Unable to decode setup")
        end
    end

    -- Set the hostname
    hostname = host

    -- At this point the "setup" is a table!

    -- These are changed in the original setup so they end up in the SIE file!
    for i, stanza in ipairs(setup) do
        -- Make sure all of the sample_rate parameters are numeric!
        for _, n in ipairs{'sample_rate', 'sample_rate_legacy_uncorrected'} do
            if setup[i][n] then
                setup[i][n] = rational:new(setup[i][n]):tonumber()
            end
        end
        -- Make sure everything that "looks" like a boolean IS a boolean
        setup[i] = fixup_booleans(stanza)
    end

    -- Get the setup into test engine order.  NOTE, a deep copy has to be made
    -- here because setup spec modifies the original!
    local _setup = tablex.deepcopy(setup)
    local preprocessed_setup =
        cjson.decode(setup_spec:new(_setup):export('test_order'))

    -- TODO: Add setup validation here!!!

    -- If this is a mode 2 networked test and a run stopper computed channel
    -- is included, don't let the test run.
    if preprocessed_setup[1].network_mode and
        tonumber(preprocessed_setup[1].network_mode) == 2 then
        for _, stanza in ipairs(preprocessed_setup) do
            if stanza.category and stanza.category == 'filter/run_stopper' then
                error('Run stopper is not compatible with ' ..
                    'mode 2 networked tests')
            end
        end
    end

    -- Get the "display" order of the channels.  Want the setup as given so
    -- a pristine copy is available!
    local display_order, dorder_sxr = get_display_order(setup)

    -- Propagate some items from summary to all other stanzas
    preprocessed_setup = propagate_summary_info(preprocessed_setup)

    -- Set prerun zero for all source stanzas that have it enabled.
    preprocessed_setup = set_prerun_zero(preprocessed_setup)

    -- Deal with "uncalibrated" channels
    preprocessed_setup = process_uncalibrated_channels(preprocessed_setup)

    -- Get the start time delay
    preprocessed_setup = get_start_time_delay(preprocessed_setup)

    -- Do notification throttling preprocessing
    preprocessed_setup = fixup_notification_throttling(preprocessed_setup)

    -- Fix up calibration
    preprocessed_setup = fixup_calibration(preprocessed_setup)

    -- For all MX channels, append the network node to the connector string
    preprocessed_setup = fixup_mx_connectors(preprocessed_setup)

    -- Insert channels that may be needed
    preprocessed_setup = add_raw_can_channels(preprocessed_setup)
    preprocessed_setup = add_raw_gps_channels(preprocessed_setup)

    -- Set network_node in each of the stanzas
    preprocessed_setup = add_network_node(preprocessed_setup)

    -- Add "fake" channels
    preprocessed_setup = add_elapsed_time_channel(preprocessed_setup)
    preprocessed_setup = add_sie_size_channel(preprocessed_setup)

    -- Add the "network_nodes" key to mode 2 networked test setup summary
    preprocessed_setup = add_network_nodes_list(preprocessed_setup)

    -- See comment on function
    preprocessed_setup = set_channel_counts(preprocessed_setup)

    -- Now go through the channels and figure out the resulting scale and
    -- offset in each stanza.
    preprocessed_setup = process_calibration(preprocessed_setup)

    -- Now go through the channels and set the timescale for
    -- channels if it isn't already set.
    preprocessed_setup = set_timescales(preprocessed_setup)

    -- Now go through the filter channels and if they don't have
    -- assorted parameters add them.
    preprocessed_setup = fixup_filters(preprocessed_setup)

    -- Deal with the MX channels that have "real_time_signal" stuff
    preprocessed_setup = process_realtime_signals(preprocessed_setup)

    -- Special processing for MX878 channels
    preprocessed_setup = process_mx878_channels(preprocessed_setup)

    -- Special processing for legacy eDAQ layer channels
    preprocessed_setup = process_legacy_edaq_layer_channels(preprocessed_setup)

    -- Finally, go through and assign SIE buffer offsets to all channels
    preprocessed_setup = assign_sie_buffer_offsets(preprocessed_setup)

    -- Do the final cleanup of the setup
    preprocessed_setup = cleanup_setup(preprocessed_setup)

    -- Do all of the mode 2 networking preprocessing
    preprocessed_setup = do_mode2_networked_processing(preprocessed_setup)

    -- All done with preprocessing, return the preprocessed setup as a table!
    return preprocessed_setup, display_order, dorder_sxr
end
