--[[
    Initial Author: Joseph Kyrouac

    Copyright (c) 2012 by Hottinger Baldwin Messetechnik Inc.
    All rights reserved.

    This source code contains proprietary and confidential
    information belonging to Hottinger Baldwin Messetechnik Inc.
    Storage, duplication, transmission, and/or disclosure of
    this information by physical, electronic, or other methods
    is prohibited except as provided under license by HBM.

    This notice does not evidence any actual or intended
    publication of this material.

    FILE:
           tablex.lua

    PURPOSE:
           This module contains various helping functions for
        tables, including a print, deepcompare and deepcopy.

    NOTES:
--]]

tablex = {}

--if object is a table, it will make a deep copy, including all subtables
function tablex.deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

function tablex.deepprint(object, indent)
    local indent = indent or 4
    if type(object) ~= "table" then
        io.write( tostring(object) .. '\n')
    else
        io.write('{\n')
        indent = indent + 4
        for k,v in pairs(object) do
            if type(k) == 'number' then
                io.write( string.rep( " ", indent) .. "[" .. k .. "] = " )
            else
                io.write( string.rep( " ", indent) .. k .. " = " )
            end
            tablex.deepprint(v, indent)
        end
        indent = indent - 4
        io.write(string.rep( " ", indent) .. '}\n')
    end
end

--will not detect loops in tables
function tablex.tostring(object, indent)
    local indent = indent or 4
    local str = ""
    local maxlen = 0
    if type(object) ~= "table" then
        str = str .. tostring(object) .. '\n'
    else
        str = str .. '{\n'
        for k,v in pairs(object) do
            if #tostring(k) > maxlen then
                maxlen = #tostring(k)
            end
            str = str .. string.rep(" ", indent) .. "[" .. k .. "] = "..
                string.rep(" ", maxlen-#tostring(k))..
                tostring(tablex.tostring(v, indent+4))
        end
        str = str .. string.rep(" ", indent-4) .. '}\n'
    end
    return str
end

--If object is a table, it will do a deep comparison, including all subtables.
function tablex.deepcompare(t1,t2,ignore_mt)
    local ty1 = type(t1)
    local ty2 = type(t2)
    if ty1 ~= ty2 then return false end
    -- non-table types can be directly compared
    if ty1 ~= 'table' and ty2 ~= 'table' then return t1 == t2 end
    -- as well as tables which have the metamethod __eq
    local mt = getmetatable(t1)
    if not ignore_mt and mt and mt.__eq then return t1 == t2 end
    for k1,v1 in pairs(t1) do
        local v2 = t2[k1]
        if v2 == nil or not tablex.deepcompare(v1,v2) then return false end
    end
    for k2,v2 in pairs(t2) do
        local v1 = t1[k2]
        if v1 == nil or not tablex.deepcompare(v1,v2) then return false end
    end
    return true
end

function tablex.num_elem(t)
    local elem = 0
    for k,v in pairs(t) do elem=elem+1 end
    return elem
end

function tablex.array_concat(a1, a2)
    if type(a1) ~= 'table' or type(a2) ~= 'table' then
        error("Invalid use of array_concat")
    end
    for i = 1, #a2 do
        a1[#a1+1] = a2[i]
    end
    return a1
end

function tablex.collect_sparse_array(input_array)
    if tablex.num_elem(input_array) > 0 then
        local max = 0
        for k,v in pairs(input_array) do
            if tonumber(k) > max then
                max = tonumber(k)
            end
        end
        local output_array = {}
        for i=0,max do
            if input_array[i] then
                output_array[#output_array+1] = input_array[i]
            end
        end
        return output_array
    end
    return input_array
end

return tablex

