require('lj_include'){
    'signal',
}

local ffi = require 'ffi'
local C = ffi.C
local cjson = require 'cjson.safe'
local ev = require 'ev'
local sqlite3 = require 'lsqlite3'
local log = require 'log'
local ts = require 'timestamps'
local uuid = require 'uuid'
local unix = require 'unix'

-- Functions used by the test control daemon for automatic SIE file upload

local m = {}

-- sxrpp - Preprocessed SXR file
-- friendly_sie_name - Name file will be called on FTP server
-- sie_name - path to "hash" named SIE file
function m:do_auto_upload(summary, friendly_sie_name, sie_name, master)
    -- If master DBUS services aren't given, just return
    if not master then return end

    -- If this is a mode2 networked test and is a slave, don't do upload
    -- because the master does that.
    if summary.is_mode2_networking and not summary.is_master_node then
        return
    end

    -- If this is not an SIE test, don't bother there is not an SIE file to
    -- upload.
    if summary.purpose then return end

    local sie_auto_upload = summary.cyclic_test_options and
        summary.cyclic_test_options.sie_auto_upload

    if not sie_auto_upload then return end

    local function get_pref_info()
        local db = sqlite3.open('/var/lib/armada/setup.db')
        if not db then return end
        local si
        local sok = false
        for r in db:rows(string.format(
            [[SELECT content FROM tsd_users WHERE db_name = 'ftp_servers' AND
            deleted = '';]])) do
            si = cjson.decode(r[1])
            if si and si.active and si.username and si.password and
                si.server_ip then
                sok = true
                break
            end
        end
        db:close()
        return sok and si
    end

    sie_auto_upload.ftp = get_pref_info()
for k, v in pairs(sie_auto_upload.ftp) do print("", k, v) end
    if sie_auto_upload and not sie_auto_upload.ftp then
        -- Auto upload is enabled but there is no "active" FTP server
        log:log(log.NOTICE,
            "FTP upload is enabled but there is no active FTP server")
--        self:send_notification_email('event', 'no_active_ftp_server')
        return
    end

    local now = ts.get_ts()

    -- Do automatic SIE upload if configured to do so
    if sie_auto_upload and sie_auto_upload.ftp and
        sie_auto_upload.ftp.username and sie_auto_upload.ftp.password and
        sie_auto_upload.ftp.server_ip then

        -- If the file naming scheme is date/time, use it as-is as the
        -- FTP filename; otherwise (numeric increment filenames), tack
        -- on a timestamp to ensure uniqueness.
        local remote_name
        if summary and summary.sie_filename_mode == 'utc_date_time' then
            remote_name = friendly_sie_name .. ".sie"
        else
            remote_name = friendly_sie_name .. "__" ..
                ts.ts_to_iso(ts.get_ts()) .. ".sie"
        end

        -- FTP doesn't like ":" in the file name so we replace them with "-"
        remote_name = string.gsub(remote_name, ':', '-')

        log:notice(
            '{0} Spawning process to upload SIE {1}.sie file to {2} now = {3}',
            summary.test_name, remote_name,
            sie_auto_upload.ftp.server_ip,
            ts.epoch_to_iso(now.s + now.ns / 1000000000))

-- TODO:
--[[
        local email_recipients = self.email_info and self.email_info.to and
            table.concat(self.email_info.to, ',')
]]
local email_recipients

        -- If a remote directory path was specified, prepend it to the remote
        -- file name
        if sie_auto_upload.ftp.destination_path and
            type(sie_auto_upload.ftp.destination_path) == 'string' and
            #sie_auto_upload.ftp.destination_path > 0 then
            if not sie_auto_upload.ftp.destination_path:find("/$") then
                sie_auto_upload.ftp.destination_path =
                    sie_auto_upload.ftp.destination_path .. '/'
            end
            remote_name = sie_auto_upload.ftp.destination_path .. remote_name
        end

        local pid = master:SpawnNonCritical(
            '/usr/bin/upload-sie',
            {
                '/usr/bin/upload-sie',
                '"' .. remote_name .. '"',      -- Name on the FTP server
                '/storage/' .. uuid.generate() .. '.sie',       -- Name of fifo
                sie_auto_upload.ftp.post_remove and 'remove' or 'keep',
                'ftp',                          -- Protocol (FTP)
                sie_auto_upload.ftp.username,   -- FTP login name
                sie_auto_upload.ftp.password,   -- FTP password
                sie_auto_upload.ftp.server_ip,  -- FTP host IP address
                email_recipients,               -- Email notifcation receipients
            })
        if not pid then
            log:log(log.ERR, string.format(
                "Unable to upload SIE file %s to %s: " ..
                "Unable to spawn upload process",
                friendly_sie_name,
                sie_auto_upload.ftp.server_ip))
        elseif not sie_auto_upload.delay then
            -- Signal the uploader to upload the SIE file immediately
            ev.Timer.new(function(l, t, e)
                    t:stop(l)
                    C.kill(ffi.new('pid_t', tonumber(pid)), unix.SIGUSR1)
                end, .5, 0):start(ev.Loop.default)
        else
            -- If automatic upload is enabled AND it is configured
            -- to upload after a delay when the next cycle starts
            -- the signal sent to the upload process will be sent
            -- after the desired delay when that cycle starts.
            ev.Timer.new(function(l, t, e)
                    t:stop(l)
                    C.kill(ffi.new('pid_t', tonumber(pid)), unix.SIGUSR1)
                end, sie_auto_upload.delay, 0):start(ev.Loop.default)
        end
    end
end

return m.do_auto_upload
