require 'strict'

-- bitmap_trigger.lua - Implements the Bitmap Trigger computed channel.

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local rational = require 'rational'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_min = math.min
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Utils_frame_data_type = utils.frame_data_type
local Utils_put_frame = utils.put_frame

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

function mt:flush_output_frame()
    self.of.count = self.bufsize
    self.of.hdr.ts = self.tot_count
    self.of.ts = math.floor(self.tot_count)
    if not Utils_put_frame(self.of, self.pipe_list) then
        Frame_delete(self.of)
        self.of = nil
        self.out_vals = nil
    end
    self.of = Frame_create(self.bufsize, self.out_frame_sample_size)
    self.out_vals = ffi_cast(Utils_frame_data_type(
        self.out_frame_sample_size), self.of.buf)
    self.tot_count = self.tot_count + self.bufsize
    self.out_index = 0
end

function mt:process_in_frame(f)
    local in_vals = ffi_cast(Utils_frame_data_type(f.element_size), f.buf)
    if     self.procmode == "any_not_inverted" then
        for i = 0, f.count-1 do
            local result = bit.band(in_vals[i],self.bitmask) ~= 0
            self.out_vals[i] = result and 1 or 0
        end
    elseif self.procmode == "all_not_inverted" then
        for i = 0, f.count-1 do
            local result = bit.band(in_vals[i],self.bitmask) == self.bitmask
            self.out_vals[i] = result and 1 or 0
        end
    elseif self.procmode == "any_inverted" then
        for i = 0, f.count-1 do
            local result = bit.band(in_vals[i],self.bitmask) == 0
            self.out_vals[i] = result and 1 or 0
        end
    elseif self.procmode == "all_inverted" then
        for i = 0, f.count-1 do
            local result = bit.band(in_vals[i],self.bitmask) == self.bitmask
            self.out_vals[i] = result and 0 or 1
        end
    end
    self:flush_output_frame()
end

-- Processes incoming frames, produces outgoing frames.
-- This is broken into many smaller functions due to the way the
--   LuaJIT compiler handles edge-cases.  To be ported to C in the future.
function mt:run()
    while Pipe_has_frames(self.in_pipe) do
        local f = Pipe_get_frame(self.in_pipe)
        self:process_in_frame(f)
        Frame_delete(f)
        f = nil
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local bittrig = setmetatable({}, {__index = mt})
    bittrig.sie_collect_idx = first_sie_index
    utils.set_common_meta(bittrig, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Bitmap Trigger computed channel " ..
        "only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    assert(input_node.is_logical ~= true, "Input channel must not be a logical "
        .."channel")
        --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    bittrig.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(bittrig.pipe_rate and bittrig.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(bittrig.pipe_rate))
        --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    bittrig.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(bittrig.sample_rate and bittrig.sample_rate > 0,
        "Invalid sample rate: " .. tostring(bittrig.sample_rate))
    bittrig.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(bittrig.out_frame_sample_size and bittrig.out_frame_sample_size > 0,
        "Invalid frame sample size:  "..tostring(bittrig.out_frame_sample_size))

    --Module specific parameters
    assert(stanza.mode, "Missing Bitmap Trigger field \"mode\"")
    assert(stanza.mode == "any" or stanza.mode == "all",
        "Bitmap Check Mode must be either \"any\" or \"all\"")
    assert(tonumber(stanza.bitmask), "Missing Bit Mask field")
    bittrig.bitmask = rational:new(stanza.bitmask):tonumber()
    assert(bittrig.bitmask >= 1 and bittrig.bitmask <= 255, "Bit Mask field"..
        " must be between 0x1 and 0xFF.")
    local invert = utils.xlate_meta_boolean(stanza.invert)
    assert(not invert or invert == true, "Inverted field "..
        "must be either true or false")
    if stanza.mode == "any" and not invert then
        bittrig.procmode = "any_not_inverted"
    elseif stanza.mode == "all" and not invert then
        bittrig.procmode = "all_not_inverted"
    elseif stanza.mode == "any" and invert then
        bittrig.procmode = "any_inverted"
    else -- stanza.mode == "all" and invert then
        bittrig.procmode = "all_inverted"
    end

    -- create input/output pipe structures
    bittrig.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(bittrig.in_pipe and bittrig.in_pipe ~= nil,
        "failed to create input pipe")
    bittrig.pipe_list = Pipe_new_pipe_list()
    assert(bittrig.pipe_list and bittrig.pipe_list ~= nil,
        "failed to create (output) pipe list")

    --local variables needed not set from stanza.
    bittrig.tot_count = 0
    bittrig.bufsize = frame.bufsize(bittrig.sample_rate, bittrig.pipe_rate)
    -- output frame
    bittrig.of = Frame_create(bittrig.bufsize, bittrig.out_frame_sample_size)
    -- "pointer" to output frame data
    bittrig.out_vals = ffi_cast(Utils_frame_data_type(
        bittrig.out_frame_sample_size), bittrig.of.buf)
    bittrig.out_index = 0

    bittrig.w_res = bittrig.factor -- # of samples needed to complete window

    return bittrig, 1
end

return m
