require 'strict'

-- state_mapper.lua - Implements the State Mapper computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof
local math_ceil = math.ceil
local math_min = math.min

local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}

function mt:dump()
    dbg('use_default_output=%s, default_output=%s, trans_win_samples=%s',
        tostring(self.use_default_output), tostring(self.default_output),
        tostring(self.trans_win_samples))
end

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Converts the given input value into the associated state value.
-- NOTE:  The states are checked in order from first to last.  This
--   allows the user to perform his/her own optimal ordering of states.
-- Parameter:  in_val - a single value from the input channel
-- Returns:  the computed state value
function mt:mapped_val(in_val)
    local val = tonumber(in_val)
    local mval
    -- set value to use if input does not fall within any
    -- state's range
    if self.use_default_output then
        mval = self.default_output
    else
        mval = self.out_value
    end

    -- Look for the first state in which the input value fits.
    for i = 1, #self.states do
        if val >= self.states[i].min and
           val < self.states[i].max then
            mval = self.states[i].val
            break
        end
    end

    return mval
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    if not Pipe_has_frames(self.in_pipe) then
        return
    end

    -- Process each frame in input pipe.  There is only one input pipe.
    local in_vals      -- "pointer" to current input frame's buf
    local of  -- current output frame
    local out_vals     -- "pointer" to current output frame's buf
    local f = Pipe_get_frame(self.in_pipe)
    while f ~= nil do
        -- create frame for output
        of, in_vals, out_vals =
            utils.create_output_frame(f.maxsize, f,
                self.tot_count, self.out_frame_sample_size)
        -- fill output frame
        if self.trans_win_samples == 0 then
            -- no "transient rejection" window, so output each
            -- mapped input
            for i = 0, f.count - 1 do
                -- set output value
                self.out_value = self:mapped_val(in_vals[i])
                out_vals[i] = self.out_value
            end
        else -- change output only after "transient rejection"
             -- window passes without further change;
             -- NOTE that window size is at least 1 sample
            local candidate = 0  -- stores current mapped value
            for i = 0, f.count - 1 do
                candidate = self:mapped_val(in_vals[i])
                if self.first_sample == true then
                    self.first_sample = false
                    self.out_value = candidate
                    self.prev_candidate = candidate
                elseif self.prev_candidate ~= candidate then
                    -- restart "transients rejection" window
                    self.trans_win_count = 0
                    self.prev_candidate = candidate
                elseif candidate ~= self.out_value then
                    self.trans_win_count = self.trans_win_count + 1
                    if self.trans_win_count >= self.trans_win_samples then
                        self.out_value = candidate
                        self.trans_win_count = 0
                    end
                end
                out_vals[i] = self.out_value
            end
        end

        of.count = f.count
        self.tot_count = self.tot_count + of.count
        -- put output frame in output pipe(s)
        if not utils.put_frame(of, self.pipe_list) then
            Frame_delete(of)
        end

        -- Delete frames that have been processed and get another set.
        -- Make sure we only get frames when all of the pipes have frames
        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    local mapper = setmetatable({}, {__index = mt})
    mapper.stanza = stanza
    mapper.sie_collect_idx = first_sie_index
    -- validate needed metadata
    assert(mapper.stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    --   input channel (exactly one)
    local input_channels = mapper.stanza.input_channels or {}
    assert(#input_channels == 1, "State Mapper computed channel only" ..
        " accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    --   sample rate
    assert(mapper.stanza.sample_rate, "Missing sample rate")
    mapper.sample_rate =
        rational:new(mapper.stanza.sample_rate):tonumber()
    assert(mapper.sample_rate and mapper.sample_rate > 0,
        "Invalid sample rate: " .. mapper.sample_rate)
    --   output frame element size
    assert(mapper.stanza.data_type, "Missing data type")
    mapper.out_frame_sample_size = 8  -- for 64-bit float
    --   state descriptions
    mapper.states = mapper.stanza.states
    assert(mapper.states, "Missing parameter 'states'")
    --   "default output" values
    if type(mapper.stanza.use_default_output) == "boolean" then
        mapper.use_default_output = mapper.stanza.use_default_output
    elseif mapper.stanza.use_default_output == "true" then
        mapper.use_default_output = true
    else
        mapper.use_default_output = false
    end
    mapper.default_output =
        rational:new(mapper.stanza.default_output):tonumber()
    assert(mapper.default_output, "Missing default output value")
    --   "transient rejection" window length
    if mapper.stanza.latch_period then
        local latch = rational:new(mapper.stanza.latch_period):tonumber()
        assert(latch >= 0, "Invalid Latch Period:  " .. latch)
        mapper.trans_win_samples =
            math.floor(latch * mapper.sample_rate + 0.5)
        mapper.trans_win_count = 0
    end
    
    -- create input/output pipe structures
    mapper.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(mapper.in_pipe and mapper.in_pipe ~= nil,
        "failed to create input pipe")
    mapper.pipe_list = Pipe_new_pipe_list()
    assert(mapper.pipe_list and mapper.pipe_list ~= nil,
        "failed to create (output) pipe list")

    -- stores most recently output state
    mapper.out_value = mapper.default_output
    -- stores the state mapped from the immediately previous input
    mapper.prev_candidate = mapper.default_output
    -- in effect only for the first sample in the test
    mapper.first_sample = true

    -- used for frames' time stamps (just a sample count for
    -- this module)
    mapper.tot_count = 0

    return mapper, 1
end

return m
