require 'strict'

-- message_logger.lua - DataMode which allows the use of triggers to
-- determine when to store message data.

local cjson = require 'cjson'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local __ = require '__'
local log = require 'log'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:
--     'sie_writer'  SIE writer
-- Returns:  true
function mt:setup(sie_writer)
    self.sw = sie_writer
    self:sie_xml()
    return true
end

-- Creates the SIE buffers (pipes and XML) required.
-- There will be one channel description per input channel.
function mt:sie_xml()
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        -- write XML for this SIE channel
        info.stanza.description = self.stanza.description or ""
        info.stanza.datamode_name = self.name
        info.stanza.sie_name = self.name .. '@' .. info.name
        info.stanza.somat_input_channel = info.stanza.name
        self.stanza.nbr_data_bits = ''
        local schema_name = sie_schema.get_collect_schema_name(self.chan_info[i])
        local schema_info = sie_schema.get_schema_info(schema_name, self.stanza)
        if self.stanza.trigger_mode == "always_on" then
            schema_info.dims[1]['core:label'] = "Time"
            schema_info.dims[1]['core:units'] = "Seconds"
        else
            schema_info.dims[1]['core:label'] = "Collection Time"
            schema_info.dims[1]['core:units'] = "Seconds"
        end
        schema_info.dims[2]['core:label'] = info.name
        schema_info.dims[2]['core:units'] = info.stanza.output_units
        local md = sie_utils.write_sie_channel(self.sw, schema_info,
            info.stanza)
        -- Attach one end of pipe to this module for this input node
        local p = Pipe_attach_pipe(info.pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        local not_fixed_size
        if info.stanza.data_type:find("message_raw") and
            (info.stanza.category:find("msg_gps$") or
             info.stanza.category:find("msg_camera$")) then
            not_fixed_size = true
        end
        local size = utils.sie_buf_size(self, info.stanza)
        self.sw:new_buffer(self.sie_collect_idx + i - 1, md._sie_group,
            size, nil, 8, p, true, nil, not_fixed_size)
    end
end

function mt:at_least_one_channel_has_frame()
    local got_a_frame = false
    for i = 1, #self.chan_info do
        got_a_frame = got_a_frame or Pipe_has_frames(self.chan_info[i].in_pipe)
        if got_a_frame then break end
    end
    return got_a_frame
end

function mt:run()
    if self.trigger_module then
        self.trigger_module:process_messages()
    end
    -- Process any frame in input pipe(s).
    while self:at_least_one_channel_has_frame() do
        for i = 1, #self.chan_info do
            local chan = self.chan_info[i]
            while Pipe_has_frames(chan.in_pipe) do
                chan.in_frame = Pipe_get_frame(chan.in_pipe)
                if not utils.put_frame(chan.in_frame, chan.pipe_list) then
                    Frame_delete(chan.in_frame)
                end
            end
        end
    end

end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer for
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
--   meta_list --> list of all created-to-this-point nodes' stanzas,
--       indexed by channel name
-- Return values:
--   the new instance object
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    assert(first_sie_index, "Missing parameter 'first_sie_index'")
    assert(meta_list, "Missing parameter 'meta_list'")
    local msg_log = setmetatable({}, {__index = mt})
    msg_log.stanza = stanza
    -- don't need to call set_common_meta() since there are
    -- no modules downstream
    msg_log.sie_collect_idx = first_sie_index
    msg_log.chan_info = {}
    -- validate and process needed metadata
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1, "Message Logger DataMode " ..
        "must have at least one input channel")
    --     node name
    msg_log.name = stanza.name
    assert(msg_log.name, "Missing node name")    
    --     trigger
    local trigger_mode = stanza.trigger_mode
    assert(trigger_mode, "Missing trigger mode")
    local trigger_module
    
    if trigger_mode ~= "always_on" then
        --     create trigger module instance
        assert(trigger_mode ~= "one_shot", "Message Logger DataMode does not "..
            "support one shot triggers")
        local trig_chan = stanza.trigger_channel
        assert(trig_chan, "Trigger mode requires a trigger channel")
        local ok, mod =
            pcall(require, 'test_engine/modules/trigger/' .. trigger_mode)
        if not ok then
            error(string.format('Could not load trigger ' ..
                 'module %s (%s)', stanza.trigger_mode, stanza.name))
        end
        for i = 1, #input_channels do
            local node_meta = meta_list[input_channels[i]]
            assert(node_meta, "metadata stanza for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            local data_type = node_meta.data_type
            assert(data_type, "Missing data_type for \"" ..
                tostring(input_channels[i]) .. "\"")
        end
        trigger_module = mod:new(input_channels, trig_chan, node_list, meta_list)
        if not trigger_module then
            error(string.format(__'Unable to create type %s (%s)',
                'test_engine/modules/trigger/' .. trigger_mode,
                stanza.name))
        end
        assert(#trigger_module.chan_info == #input_channels, "Number of "..
            "trigger output pipes does not equal number of input channels")
        msg_log.trigger_module = trigger_module
    end
    
    for i = 1, #input_channels do
        local node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        assert(node_meta.data_type, "Missing data_type for \"" ..
            tostring(input_channels[i]) .. "\"")
        assert(node_meta.data_type == "message_raw" or
            node_meta.data_type == "message_can",
            "Message Logger may only store raw message or raw CAN channels.")
        msg_log.chan_info[i] =
            {name = input_channels[i],
             pipe_list = Pipe_new_pipe_list(),
             stanza = node_meta,
             data_type = node_meta.data_type}
        if trigger_mode == "always_on" then
            -- Create input pipe structures
            local input_node = node_list[input_channels[i]]
            assert(input_node, "Node for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            msg_log.chan_info[i].in_pipe = 
                Pipe_attach_pipe(input_node.pipe_list)
        else
            -- Create input pipe structures using trigger module output pipes
            -- as this instance's input pipes
            msg_log.chan_info[i].in_pipe = Pipe_attach_pipe(
                 trigger_module.chan_info[i].pipe_list)
        end
        assert(msg_log.chan_info[i].in_pipe and
            msg_log.chan_info[i].in_pipe ~= nil,
            "could not create input pipe")
        assert(msg_log.chan_info[i].pipe_list and
            msg_log.chan_info[i].pipe_list ~= nil,
            "could not create pipe list")
    end
    return msg_log, #input_channels  -- 1 sie buf per input chan
end

return m
