require 'strict'

-- peak_valley_slice.lua - Implements the Peak Valley Slice data mode.
--
local cjson = require 'cjson'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local __ = require '__'
local log = require 'log'
local pipe = require 'test_engine/pipe'
local pv = require 'test_engine/modules/common/pv_slice_algo'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}


-- Performs any needed post instantiation initialization.
-- Parameter:
--     'sie_writer'  SIE writer
-- Returns:  true
function mt:setup(sie_writer)
    self.sw = sie_writer
    self:sie_xml()
    return true
end

-- Creates the SIE buffers (pipes and XML) required.
-- There will be one channel description per input channel.
function mt:sie_xml()
    for i = 1, self.nbr_output_chans do
        local info = self.chan_info[i]
        -- write XML for this SIE channel
        info.stanza.description = self.stanza.description or ""
        info.stanza.datamode_name = self.name
        info.stanza.sie_name = self.name .. '@' .. info.name
        info.stanza.somat_input_channel = info.stanza.name
        -- the following line makes the schema info reflect the
        -- input channel's data bit size (?should this be different?)
        self.stanza.nbr_data_bits =
            tonumber(info.data_type:match('^.-(%d+)$'))
        assert(self.stanza.nbr_data_bits and self.stanza.nbr_data_bits > 0,
            "Missing data bit size")
        local schema_name = sie_schema.get_dm_schema_name(self.stanza.category)
        local schema_info =
            sie_schema.get_schema_info(schema_name, self.stanza)
        schema_info.somat_datamode_type = 'peak_valley_slice'
        schema_info.dims[1]['core:label'] = "PV Sequence"
        schema_info.dims[1]['core:units'] = nil
        schema_info.dims[2]['core:label'] = info.stanza.name
        schema_info.dims[2]['core:units'] = info.stanza.output_units
        schema_info.xform_scale = 1.0
        schema_info['core:uuid'] = self.stanza.id .. '/' .. info.stanza.id
        local md = sie_utils.write_sie_channel(self.sw, schema_info,
            info.stanza)
        -- Attach one end of pipe to this module for this input node
        local p = Pipe_attach_pipe(info.pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        local size = utils.sie_buf_size(self, info.stanza)
        self.sw:new_buffer(self.sie_collect_idx + i - 1,
            md._sie_group, size,
            ffi_sizeof(info.data_type), 8, p)
    end
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mt:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = false

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame =
                Pipe_get_frame(self.chan_info[i].in_pipe)
        end
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
    end
    return got_frames
end

-- Update the given array of slices by adding the slices in
-- 'slice_table' and putting all slices in proper timestamp order
-- in the class instance variable 'pvs_array'.
-- Params:  'slice_table'  an array of slice arrays (see below)
--
-- Data structures:
--      A "slice" is a table with two elements:
--          { 'vals' --> array of sample values, 'ts' --> timestamp }
--      "slices" is an array of "slice" tables
--      'slice_table' is an array of "slices", indexed by the
--          master channel index
--      NOTE that each of the arrays are indexed from 1.
function mt:update_slice_list(slice_table)
    if #slice_table > 0 then
        if self.pvs_array == nil then
            self.pvs_array = {}
        end
        -- append all slices in slice_table to array
        for c = 1, self.nbr_master_chans do
            for _, slice in ipairs(slice_table[c]) do
                assert(type(slice) == "table",
                       "INTERNAL ERROR: non-table slice")
                self.pvs_array[#self.pvs_array + 1] = slice
            end
        end
        -- sort array in ascending timestamp order
        table.sort(self.pvs_array, function(a, b)
                return a.ts < b.ts and true or false
            end)
        -- remove slices with duplicate timestamps
        local si = 1
        local count = #self.pvs_array
        while si < count do
            if self.pvs_array[si].ts == self.pvs_array[si + 1].ts then
                table.remove(self.pvs_array, si)
                count = #self.pvs_array
            else
                si = si + 1
            end
        end
    end
end


-- Returns a table containing all the master channels' candidate
-- slices.  The master channels' candidate references are set to nil.
-- Returns:  table of arrays of slices, with no more than one slice
--     in each array
function mt:get_currents()
    local curr_table = {}
    local candidate = nil
    local info

    for c = 1, self.nbr_master_chans do
        curr_table[c] = {}
        info = self.chan_info[c]
        if info.state == "initial" then
            candidate = info.first_valley_candidate
            info.first_valley_candidate = nil
            if candidate == nil or
               (info.first_peak_candidate ~= nil and
               candidate.ts < info.first_peak_candidate.ts) then
                candidate = info.first_peak_candidate
            end
            info.first_peak_candidate = nil
        else
            candidate = info.candidate
        end
        info.candidate = nil
        if candidate ~= nil then
            table.insert(curr_table[c], candidate)
        end
    end

    return curr_table
end

-- Puts each channel's output frames to its output pipe list.
function mt:output_frames()
    for ch_index = 1, self.nbr_output_chans do
        local info = self.chan_info[ch_index]
        info.out_frame.count = self.out_index
        if not utils.put_frame(info.out_frame,
                               info.pipe_list) then
            Frame_delete(info.out_frame)
        end
        info.out_frame = nil
    end
    self.tot_out_sample_count = self.tot_out_sample_count + self.out_index
    self.out_index = 0
end

-- Processes incoming frames, producing outgoing frames.
function mt:run()
    if self.trigger_module then
        -- trigger node feeds this node's input channels
        self.trigger_module:run()
    end

    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- don't have a full set of input frames yet
        return
    end

    -- Process each frame in input pipe(s).
    while have_all_frames do
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        It doesn't matter which input frame we use for
        --        obtaining frame count.
        local in_count = self.chan_info[1].in_frame.count

        -- Get input frame pointers.
        for ch_index = 1, #self.chan_info do
            local info = self.chan_info[ch_index]
            if not info.sample_type then
                info.element_size = info.in_frame.element_size
                info.sample_type =
                    utils.frame_data_type(info.element_size)
            end
            info.invals = ffi_cast(info.sample_type, info.in_frame.buf)
        end

        -- Check each master input frame for peaks and valleys,
        -- making slices from both master and slave inputs when a
        -- peak or valley is found.
        local slice_table =
            pv.run_pv_slice_algo(in_count, self.chan_info,
                self.nbr_master_chans, self.nbr_output_chans,
                self.hyster_array,
                tonumber(self.chan_info[1].in_frame.hdr.ts))

        -- update the list of slices to be output, putting them
        -- in proper timestamp order in self.pvs_array
        self:update_slice_list(slice_table)

        local oldest_candidate
        if self.pvs_array and #self.pvs_array > 0 then
            -- Put each master and slave channel's slice sample into
            -- an output frame.
            local slice_count = #self.pvs_array
            local slice = self.pvs_array[1]
            while slice and slice_count > 0 do
                -- Note that we will always remove the
                -- first slice from the array after we
                -- have successfully used it

                -- ensure there are no blocking candidates for
                -- this slice
                oldest_candidate =
                    pv.get_oldest_candidate(self.chan_info,
                                            self.nbr_master_chans)
                if oldest_candidate and
                   oldest_candidate.ts < slice.ts then
                    if slice_count > self.plateau_size then
                        -- treat the blocking candidate slice as
                        -- an actual slice
                        table.insert(self.pvs_array, 1, oldest_candidate)
                        slice_count = #self.pvs_array
                        slice = oldest_candidate
                        -- remove oldest candidate from its master
                        -- channel's info
                        pv.remove_candidates_at_ts(self.chan_info,
                            self.nbr_master_chans, slice.ts)
                    else
                        if self.out_index >= 1 then
                            self:output_frames()
                        end
                        break  -- cannot output any more slices for now
                    end
                end
                -- Output frames are allocated as needed.
                -- Note that output frames are allocated (and
                -- output) as a complete set.
                if self.chan_info[1].out_frame == nil then
                    local frame_count = slice_count
                    for ch_index = 1, self.nbr_output_chans do
                        local info = self.chan_info[ch_index]
                        info.out_frame =
                            Frame_create(frame_count, info.element_size)
                        info.out_frame.hdr.ts = self.tot_out_sample_count
                        info.outvals = ffi_cast(info.sample_type,
                                                info.out_frame.buf)
                    end
                end
                -- copy slice samples for output
                for c = 1, self.nbr_output_chans do
                     self.chan_info[c].outvals[self.out_index] =
                         slice.vals[c]
                end
                self.out_index = self.out_index + 1
                -- get next slice
                table.remove(self.pvs_array, 1)
                slice_count = #self.pvs_array
                if slice_count > 0 then
                    slice = self.pvs_array[1]
                else
                    slice = nil
                    self.pvs_array = nil
                end
            end
            -- check for need to output
            if self.chan_info[1].out_frame and
               (self.out_index >= self.chan_info[1].out_frame.maxsize) then
                -- output frames full
                self:output_frames()
            end
        end

        -- Keep tot_count up-to-date as a count of total input
        -- samples seen per channel.
        self.tot_count = self.tot_count + in_count

        -- don't need the input frame set any longer
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end

        -- get a new set of input frames to process
        have_all_frames = self:get_frames()
    end
end

-- Finishes processing of outgoing frames at end of test.
-- NOTE:  Any last data to be written to the SIE file must
--        be written here, since no module's run() function
--        will be called again!
-- Parameter:
--     'sie_writer'  SIE writer
function mt:finalize(sie_writer)
    -- if output frames weren't output yet, do so
    if self.out_index >= 1 then
        self:output_frames()
    end

    -- get a table of arrays of the master channels'
    -- candidate slices (there will only be one slice at most
    -- in each array)
    local candidates_table = self:get_currents()
    -- add candidate slices to list of slices to be output,
    -- putting them in self.pvs_array in proper timestamp order
    self:update_slice_list(candidates_table)

    -- output all slices which haven't been output yet
    if self.pvs_array and #self.pvs_array > 0 then
        local slice_count = #self.pvs_array
        local slice = self.pvs_array[1]
        while slice do -- Note that we will always remove the
                       -- first slice from the array after we
                       -- have successfully used it
            if self.chan_info[1].out_frame == nil then
                for ch_index = 1, self.nbr_output_chans do
                    local info = self.chan_info[ch_index]
                    local frame_count = slice_count
                    info.out_frame =
                        Frame_create(frame_count, info.element_size)
                    info.out_frame.hdr.ts = self.tot_out_sample_count
                    info.outvals = ffi_cast(info.sample_type,
                                            info.out_frame.buf)
                end
            end
            -- copy slice samples for output
            for c = 1, self.nbr_output_chans do
                 self.chan_info[c].outvals[self.out_index] = slice.vals[c]
            end
            self.out_index = self.out_index + 1
            -- get next slice
            table.remove(self.pvs_array, 1)
            slice_count = #self.pvs_array
            if slice_count > 0 then
                slice = self.pvs_array[1]
            else
                slice = nil
            end
        end
        if self.out_index >= 1 then
            -- output frames containing remaining slices
            self:output_frames()
            self.sw:flush_these(self.sie_collect_idx,
                                self.nbr_output_chans,
                                true)  -- forced
        end
    end
end

local m = {}

-- Creates a new instantiation of this datamode type.
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer for
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
--   meta_list --> list of all created-to-this-point nodes' stanzas,
--       indexed by channel name
-- Return values:
--   the new instance object
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    assert(first_sie_index, "Missing parameter 'first_sie_index'")
    assert(meta_list, "Missing parameter 'meta_list'")
    local pvs_obj = setmetatable({}, {__index = mt})
    pvs_obj.stanza = stanza
    -- don't need to call set_common_meta() since there are
    -- no modules downstream
    pvs_obj.sie_collect_idx = first_sie_index
    pvs_obj.chan_info = {}
    -- validate and process needed metadata
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1, "Peak Valley Slice data mode " ..
        "must have at least 1 input channel")
    local master_channels = stanza.master_input_channels
    assert(#master_channels >= 1, "Peak Valley Slice data mode " ..
        "must have at least 1 master channel")
    pvs_obj.nbr_master_chans = #master_channels
    local slave_channels = stanza.slave_input_channels or {}
    -- Verify that channels are not duplicated in within the
    -- the tables input_channels, master_channels, and
    -- slave_channels.
    local sorted_input_channels = {}
    for i = 1, #input_channels do
        sorted_input_channels[i] = input_channels[i]
    end
    table.sort(sorted_input_channels)
    for i = 1, #input_channels - 1 do
        assert(sorted_input_channels[i] ~= sorted_input_channels[i + 1],
            "Duplicate channels not allowed in 'input_channels'")
    end
    local sorted_master_channels = {}
    for i = 1, #master_channels do
        sorted_master_channels[i] = master_channels[i]
    end
    table.sort(sorted_master_channels)
    for i = 1, #master_channels - 1 do
        assert(sorted_master_channels[i] ~= sorted_master_channels[i + 1],
            "Duplicate channels not allowed in 'master_channels'")
    end
    local sorted_slave_channels = {}
    for i = 1, #slave_channels do
        sorted_slave_channels[i] = slave_channels[i]
    end
    table.sort(sorted_slave_channels)
    for i = 1, #slave_channels - 1 do
        assert(sorted_slave_channels[i] ~= sorted_slave_channels[i + 1],
            "Duplicate channels not allowed in 'slave_channels'")
    end
    pvs_obj.nbr_output_chans = #master_channels + #slave_channels
    assert(pvs_obj.nbr_output_chans == #input_channels,
        "Peak Valley Slice data mode input channels list must be " ..
        "same size as master channel list plus slave channel list")
    --     verify that a channel appears exactly once in
    --     input_channels table and exactly once in either
    --     the master_channels or slave_channels table.
    local sm_index, ss_index = 1, 1
    for i = 1, #input_channels do
        if sorted_input_channels[i] == sorted_master_channels[sm_index] then
            sm_index = sm_index + 1
        elseif sorted_input_channels[i] == sorted_slave_channels[ss_index] then
            ss_index = ss_index + 1
        else
            assert(false, string.format("input channel '%s' " ..
                "not listed in master or slave channel lists",
                tostring(sorted_input_channels[i])))
        end
    end
    --     Plateau Size
    pvs_obj.plateau_size = tonumber(stanza.plateau_size)
    assert(pvs_obj.plateau_size and (pvs_obj.plateau_size >= 50) and
           (pvs_obj.plateau_size <= 500), "Missing or invalid plateau_size")
    --     node name
    pvs_obj.name = stanza.name
    assert(pvs_obj.name, "Missing node name")
    --     combine master channel(s) and slave channel(s) names
    --     into one list with the master channel(s) first
    pvs_obj.all_input_chans = {}
    for i = 1, #master_channels do
        pvs_obj.all_input_chans[i] = master_channels[i]
    end
    for i = 1, #slave_channels do
        pvs_obj.all_input_chans[#pvs_obj.all_input_chans + 1] =
            slave_channels[i]
    end
    --     sample rate (get and check for consistency)
    for i = 1, #pvs_obj.all_input_chans do
        local node_meta = meta_list[pvs_obj.all_input_chans[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(pvs_obj.all_input_chans[i]) .. "\" not found")
        if not pvs_obj.sample_rate then
            pvs_obj.sample_rate = node_meta.sample_rate
        else
            assert(pvs_obj.sample_rate == node_meta.sample_rate,
                tostring(pvs_obj.all_input_chans[i]) .. " sample rate " ..
                "does not match")
        end
    end
    pvs_obj.sample_rate = rational:new(pvs_obj.sample_rate):tonumber()
    assert(pvs_obj.sample_rate and pvs_obj.sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(pvs_obj.sample_rate)))
    --     trigger
    local trigger_mode = stanza.trigger_mode
    assert(trigger_mode, "Missing trigger mode")
    --     Table to hold hysteresis values for master channels
    pvs_obj.hyster_array = {}
    --     Hysteresis value
    assert(stanza.pv_hysteresis, "Missing pv_hysteresis value")
    local hyster_val, hyster_percent =
        tostring(stanza.pv_hysteresis):match("([%d%.]+)(%%?)")
    hyster_val = rational:new(hyster_val):tonumber()
    if hyster_percent == "%" then
        hyster_percent = hyster_val
        assert((type(hyster_percent) == 'number') and
               (hyster_percent >= 0.0) and (hyster_percent <= 100.0),
               "Invalid hysteresis percent")
        hyster_percent = hyster_percent / 100.0
    else
        hyster_percent = nil
        assert((type(hyster_val) == 'number') and (hyster_val >= 0.0),
               "Invalid hysteresis value")
    end
    --     data_type
    --     (create master input channel info objects to hold
    --      data_type and other channel-specific values)
    for i = 1, #master_channels do
        local node_meta = meta_list[master_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(master_channels[i]) .. "\" not found")
        assert(node_meta.data_type, "Missing data_type for \"" ..
            tostring(master_channels[i]) .. "\"")
        if hyster_percent then
            local range_min = tonumber(node_meta.range_min)
            assert(type(range_min) == "number",
                   "Missing or invalid range_min for \"" ..
                   tostring(master_channels[i]) .. "\"")
            local range_max = tonumber(node_meta.range_max)
            assert(type(range_max) == "number",
                   "Missing or invalid range_max for \"" ..
                   tostring(master_channels[i]) .. "\"")
            assert(range_max > range_min)
            pvs_obj.hyster_array[i] = (range_max - range_min) * hyster_percent
        else
            pvs_obj.hyster_array[i] = hyster_val
        end
        -- Set this channel's attributes
        pvs_obj.chan_info[i] =
            {name = master_channels[i],
             stanza = node_meta,
             data_type = node_meta.data_type,
             element_size = nil,
             sample_type = nil,
             state = "initial",
             valley = math.huge,
             peak = -(math.huge - 1),
             candidate = nil,
             first_valley_candidate = nil,
             first_peak_candidate = nil,
             last_input_value = nil}
    end
    --     (create slave input channel info objects to hold
    --      data_type and other channel-specific values)
    for i = 1, #slave_channels do
        local node_meta = meta_list[slave_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(slave_channels[i]) .. "\" not found")
        assert(node_meta.data_type, "Missing data_type for \"" ..
            tostring(slave_channels[i]) .. "\"")
        -- Set this channel's attributes
        pvs_obj.chan_info[#master_channels + i] =
            {name = slave_channels[i],
             stanza = node_meta,
             data_type = node_meta.data_type,
             element_size = nil,
             sample_type = nil,
             last_input_value = nil}
    end
    -- create input pipe structures
    if trigger_mode == "always_on" then
        --     create non-trigger input pipe structures
        local input_node
        local node_meta
        for i = 1, #master_channels do
            input_node = node_list[master_channels[i]]
            assert(input_node, "Node for Input Channel \"" ..
                tostring(master_channels[i]) .. "\" not found")
            pvs_obj.chan_info[i].in_pipe =
                 Pipe_attach_pipe(input_node.pipe_list)
            assert(pvs_obj.chan_info[i].in_pipe and
                pvs_obj.chan_info[i].in_pipe ~= nil,
                "could not create input pipe")
            pvs_obj.chan_info[i].pipe_list = Pipe_new_pipe_list()
            assert(pvs_obj.chan_info[i].pipe_list and
                pvs_obj.chan_info[i].pipe_list ~= nil,
                "could not create pipe list")
        end
        for i = 1, #slave_channels do
            input_node = node_list[slave_channels[i]]
            assert(input_node, "Node for Input Channel \"" ..
                tostring(slave_channels[i]) .. "\" not found")
            pvs_obj.chan_info[#master_channels + i].in_pipe =
                 Pipe_attach_pipe(input_node.pipe_list)
            assert(pvs_obj.chan_info[#master_channels + i].in_pipe and
                pvs_obj.chan_info[#master_channels + i].in_pipe ~= nil,
                "could not create input pipe")
            pvs_obj.chan_info[#master_channels + i].pipe_list =
                Pipe_new_pipe_list()
            assert(pvs_obj.chan_info[#master_channels + i].pipe_list and
                pvs_obj.chan_info[#master_channels + i].pipe_list ~= nil,
                "could not create pipe list")
        end
    else
        --     create trigger module instance
        local trig_chan = stanza.trigger_channel
        assert(trig_chan, "Trigger mode requires a trigger channel")
        local ok, mod =
            pcall(require, 'test_engine/modules/trigger/' .. trigger_mode)
        if not ok then
            error(string.format(__'Could not load trigger ' ..
                 'module %s (%s)', stanza.trigger_mode, stanza.name))
        end
        local trigger_module =
            mod:new(pvs_obj.all_input_chans, trig_chan, node_list, meta_list)
        if not trigger_module then
            error(string.format(__'Unable to create type %s (%s)',
                'test_engine/modules/trigger/' .. trigger_mode,
                stanza.name))
        end
        assert(#trigger_module.chan_info ==
               #pvs_obj.all_input_chans, "Number of trigger output pipes " ..
               "does not equal number of input channels")
        --     create input pipe structures using
        --     trigger module output pipes as this instance's
        --     input pipes
        for i = 1, #master_channels do
            pvs_obj.chan_info[i].in_pipe =
                Pipe_attach_pipe(trigger_module.chan_info[i].pipe_list)
            assert(pvs_obj.chan_info[i].in_pipe and
                pvs_obj.chan_info[i].in_pipe ~= nil,
                "could not create input pipe")
            pvs_obj.chan_info[i].pipe_list = Pipe_new_pipe_list()
            assert(pvs_obj.chan_info[i].pipe_list and
                pvs_obj.chan_info[i].pipe_list ~= nil,
                "could not create pipe list")
        end
        for i = 1, #slave_channels do
            pvs_obj.chan_info[#master_channels + i].in_pipe =
                Pipe_attach_pipe(
                    trigger_module.chan_info[#master_channels + i].pipe_list)
            assert(pvs_obj.chan_info[#master_channels + i].in_pipe and
                pvs_obj.chan_info[#master_channels + i].in_pipe ~= nil,
                "could not create input pipe")
            pvs_obj.chan_info[#master_channels + i].pipe_list =
                Pipe_new_pipe_list()
            assert(pvs_obj.chan_info[#master_channels + i].pipe_list and
                pvs_obj.chan_info[#master_channels + i].pipe_list ~= nil,
                "could not create pipe list")
        end
        pvs_obj.trigger_module = trigger_module
    end

    -- initialize module-level volatile values
    pvs_obj.out_index = 0
    pvs_obj.tot_out_sample_count = 0
    pvs_obj.tot_count = 0

    -- second returned value will indicate one sie buf per input
    -- channel
    return pvs_obj, pvs_obj.nbr_output_chans
end

return m

