-- Test functions used for notifications (email and push)
local cjson = require 'cjson'
local ts = require 'timestamps'
local net_db = require 'network_db'
local config = require 'system_config'
local log = require 'log'

local storage_path = config.storage.path or '/storage'

local m = {}

function m:send_notification_email(event, reason, other_body)

    -- If this is a mode 2 networked test AND this is a slave node, don't send
    -- notifications.  Only the master node in mode 2 networked tests sends
    -- push or email notifications.
    if (self.summary.is_master_node or self.summary.master_node) and
        (self.summary.master_node and
            self.summary.master_node ~= self.hostname) then
        return
    end

    -- If no email info or push notifications, don't do anything
    if not self.mail and not self.pushn then return end

    if not self.email_notifications and not self.email_info then
        -- First time grab the latest receipient information from the database
        local db = sqlite3.open('/var/lib/armada/setup.db')
        if not db then return end
        for r in db:rows(string.format(
            [[SELECT content FROM tsd_users WHERE
                db_name = 'email_notifications' AND deleted = '';]])) do
            local si = cjson.decode(r[1])
            if si and si.notifications and
                type(si.notifications) == 'table' then
                self.email_notifications = si.notifications
            end
            if si and si.emails and type(si.emails) == 'table' and
                si.emails.to and type(si.emails.to) == 'table' and
                #si.emails.to > 0 then
                self.email_info = si.emails
                break
            end
        end
        db:close()
    end

    -- If notification is not enabled don't do anything
    if (not self.email_notifications or not self.email_notifications[reason])
       and not other_body then
        return
    end

    -- If push notifications are enabled but email notifications are not fake
    -- the stuff so the notification will be sent.
    if self.pushn and not self.email_info then
        self.email_info_is_temporary = true
        self.email_info = {}
    end

    local cyclic = 'Test'
    local is_cyclic
    local is_first_cycle
    local cnum = ''
    if self.summary.test_run_mode and
        self.summary.test_run_mode == 'cyclic' then
        cyclic = 'Cyclic test'
        is_cyclic = true
        if self.cycle_number then
            if self.cycle_number == 1 then
                is_first_cycle = true
            else
                is_first_cycle = false
                cnum = tostring(self.cycle_number)
            end
        else
            local f = io.open(storage_path .. '/.cyclic-test.info', 'r')
            if f then
                is_first_cycle = false
                cnum = cjson.decode(f:read('*a'))
                f:close()
                cnum = tostring(cnum.test_run_cycle) or '?'
            else
                is_first_cycle = true
            end
        end
    end

    if event == 'start' then
        if is_cyclic then
            if is_first_cycle then
                if self.summary.restart_count then
                    reason = ' first cycle due to error restart'
                else
                    reason = ' first cycle'
                end
            else
                if self.summary.restart_count then
                    reason = ' cycle ' .. cnum .. ' due to error restart'
                else
                    reason = ' cycle ' .. cnum
                end
            end
        else
            reason = ' manually or as a result of being restarted ' ..
                'after a reboot/power cycle while the test was running'
            if self.summary.restart_count then
                reason = ' as a result of restart due to error'
            end
        end
        self.email_info.subject =
            string.format("%s '%s' started", cyclic, self.test_name)
        self.email_info.body =
            string.format("%s '%s' started%s at %s\n", cyclic,
                self.test_name, reason, ts.ts_to_iso(ts.get_ts()))
        self.email_info.noblock = true
    elseif event == 'stop' then
        if not reason then return end
        local how = {
            error = 'error',
            reboot_poweroff = 'stopped due to system reboot or power off',
            user_request = 'stopped at user request',
            cycle_complete = 'stopped because cycle ' .. cnum ..  ' completed',
            out_of_storage =
                'permanently stopped because the system is out of storage space',
            cycle_limit_reached =
                'permanently stopped because the cycle limit has been reached',
        }
        self.email_info.subject =
            string.format("%s '%s' stopped", cyclic, self.test_name)
        self.email_info.body =
            string.format("%s '%s' %s at %s\n", cyclic,
                self.test_name, how[reason], ts.ts_to_iso(ts.get_ts()))
    elseif event == 'signal' then
    elseif event == 'event' then
        local events = {
            no_active_ftp_server =
                "FTP upload is enabled but there is no active FTP server.",
        }
        self.email_info.subject =
            string.format("'%s' event", self.test_name)
        self.email_info.body =
            string.format("%s '%s' event at %s - %s\n", cyclic,
                self.test_name, ts.ts_to_iso(ts.get_ts()),
                events[reason] or reason)
        self.email_info.noblock = true
    elseif event == '_other_' then
        self.email_info.subject =
            string.format("'%s'", self.test_name)
        self.email_info.body =
            string.format("%s '%s' event at %s - %s\n", cyclic,
                self.test_name, ts.ts_to_iso(ts.get_ts()), other_body or "")
        self.email_info.noblock = true
    else
        return
    end
    if self.mail then
        local ok, err = pcall(self.mail.send, self.mail, self.email_info)
        if not ok then
            log:notice("Email send failure\n" .. err)
        end
        self.email_info.noblock = nil
    end
    if self.pushn then
        local ok, json, err =
            pcall(self.pushn.send, self.pushn, self.email_info)
        if not ok or err then
            if type(json) == 'table' and json.Message then
                json = json.Message
            elseif type(json) ~= 'string' then
                json = nil
            end
            log:notice("Push notification send failure\n" .. json or '')
        end
    end

    if self.email_info_is_temporary then
        self.email_info_is_temprary = nil
        self.email_info = nil
    end
end

return m
