require 'strict'
local cjson = require 'cjson.safe'
local ffi = require 'ffi'
local net = require 'net'
local membuf = require 'membuf'
local sample_clock = require 'sample_clock'
local sample_types = require 'sample_types'
local message_can = require 'message_can'
local Frame = require 'test_engine/frame'
local Pipe = require 'test_engine/pipe'
local ev = require 'ev'
local C = ffi.C
local log = require 'log'
local config = require 'system_config'
local libte = ffi.load(config.test_engine_library)


ffi.cdef[[
    struct sockaddr_un {
        sa_family_t     sun_family;
        char            sun_path[108];
    };
    char *strerror(int errnum);

    struct _can_signal_t {
        char *                  channel;
        int                     lshift;
        int                     rshift;
        void *                  extract;
        int                     mode_lshift;
        int                     mode_rshift;
        void *                  extract_mode;
        uint32_t                mode_value;
        double                  scale;
        double                  offset;
        double                  rezero_offset;
//        void                    (*rezero_offset_cb)(struct _can_signal_t *);
        const char              *rezero_offset_cb;
        void *                  sequentializer;
        struct pipe *           pipe_list[1];
        int                     sie_idx;
        int                     frame_sample_size;
        volatile uint32_t       state;
    };

    struct can_signal {
        struct can_signal           *next;
        char                        *stanza;
        char                        *channel;
        int                         frame_sample_size;
        int                         sie_idx;
        int                         rshift_bits;
        int                         lshift_bits;
        double                      scale;
        double                      offset;
        double                      rezero_offset;
//        void                        (*rezero_offset_cb)(struct can_signal *);
        const char              *rezero_offset_cb;
        double                      db_bit_length;
        double                      db_bit_start;
        uint32_t                    my_id;
        void                        *sequentializer;
        double                      (*dh)(struct can_signal *, void *);
        struct pipe                 *pipe_list[1];
        void                        *pn;
        uint32_t                    my_mask;
        volatile uint32_t           state;
    };

    struct gps_signal {
        struct gps_signal           *next;
        char                        *stanza;
        char                        *channel;
        int                         frame_sample_size;
        int                         sie_idx;
        double                      scale;
        double                      offset;
        double                      rezero_offset;
//        void                        (*rezero_offset_cb)(struct gps_signal *);
        const char              *rezero_offset_cb;
        uint32_t                    my_id;
        void                        *sequentializer;
        void                        *extractor;
        struct pipe                 *pipe_list[1];
        struct proxy_buffer         *pn;
        char                        *gps_type;
        volatile uint32_t           state;
    };


    enum {
        PN_STATE_INIT,
        PN_STATE_CONFGURING,
        PN_STATE_READY,
        PN_STATE_ERROR
    };

    typedef struct proxy_buffer {
        struct membuf   inbuf;
        struct frame    *current_frame;
        struct pipe     *pipe_list[1];
        volatile int    state;
        int             i;
        int             sample_size;
        int             frame_sample_size;
        int             frame_elements;
        int             timestamped_samples;
        int             fd;
        ssize_t         already_read;
        int             failed;
        int             read_errno;
        ssize_t         read_ret;
        timestamp       timestamp[1];
        char            *channel;
        char            buf_type[128];
        int             sie_idx;
        // This is for scale and offset performed the test engine
        double          scale;
        double          offset;
        double          rezero_offset;
//        void            (*rezero_offset_cb)(struct proxy_buffer *);
        const char              *rezero_offset_cb;
        // The stuff below this point is for sequentializing messages
        sample_clock    clock;
        double          value;
        double          stale_value;
        timestamp       last_stamp;
        int64_t         stale_clocks;
        int64_t         stale_after;
        int64_t         sample_count;
        int             is_logical;
        void            (*byte_swapper)(struct proxy_buffer *, struct frame *);
        void            (*scale_offset_fn)(struct proxy_buffer *,
                            struct frame *);
        void            (*reader_fn)(struct proxy_buffer *);
        void            (*operating_reader_fn)(struct proxy_buffer *);
        struct can_signal       *cs_list;
        struct _can_set_t       *can_set;
        struct gps_signal       *gs_list;
        timestamp       start_timestamp;
        void            *hb_timer;
        void            *user_data;
        void            *w;
    } proxy_buffer_t;


    struct proxy_buffer *proxyNode_Create(
        const char *stanza, int fd, double pipe_rate);

    void proxyNode_SetStartTime(
        struct proxy_buffer *self, const char *iso);

    void proxyNode_Stop(struct proxy_buffer *self);
]]

local m = {}

function m:new(loop, md, host, port, ip, pipe_rate)

    if md.purpose then md.do_prerun_zero = nil end

    if md.sample_rate then md.sample_rate = tonumber(md.sample_rate) end

    local pn

    -- If the connection is remote, use IP otherwise use UNIX domain socket
    local sock = net.socket{
        family = C.AF_UNIX,
        type = C.SOCK_STREAM,
        flags = 0}
    local sock2 = net.socket{
        family = C.AF_UNIX,
        type = C.SOCK_STREAM,
        flags = 0}

    -- Setup for connect via UNIX domain socket
    local sa = ffi.new('struct sockaddr_un[1]')
    sa[0].sun_family = C.AF_UNIX
    local name = md.connector:match('^.*%@(.+)$')
    ffi.copy(sa[0].sun_path, '/tmp/proxy/' .. name)

    sock.addrs = ffi.new('struct addrinfo')
    sock.addrs.ai_flags = 0
    sock.addrs.ai_family = C.AF_UNIX
    sock.addrs.ai_socktype = C.SOCK_STREAM
    sock.addrs.ai_protocol = 0
    sock.addrs.ai_addrlen = ffi.sizeof('struct sockaddr_un')
    sock.addrs.ai_addr = ffi.cast('struct sockaddr *', sa)
    sock.addrs.ai_next = nil

    sock.sock = -1

    sock2.addrs = ffi.new('struct addrinfo')
    sock2.addrs.ai_flags = 0
    sock2.addrs.ai_family = C.AF_UNIX
    sock2.addrs.ai_socktype = C.SOCK_STREAM
    sock2.addrs.ai_protocol = 0
    sock2.addrs.ai_addrlen = ffi.sizeof('struct sockaddr_un')
    sock2.addrs.ai_addr = ffi.cast('struct sockaddr *', sa)
    sock2.addrs.ai_next = nil

    sock2.sock = -1

    -- Function used to write the channel setup information the proxy
    local function start(new_md)
        if sock.sock < 0 then
            error(string.format("Connection for channel %s is not complete",
                new_md.name), 0)
            return
        end
        libte.proxyNode_SetStartTime(pn, new_md.start_time)
        local out_md = cjson.encode(new_md)
        local msg = string.format("POST /channel_stream HTTP/1.0\r\n" ..
            "Content-length: %d\r\n\r\n%s", #out_md, out_md)

        ev.IO.new(function(l, t, e)
                local written = C.write(sock.sock, msg, #msg)
                if written > 0 then
                    msg = string.sub(msg, written + 1)
                    if #msg == 0 then
                        t:stop(l)
                    end
                elseif written == 0 then
                    t:stop(l)
                    error(string.format("Proxy for channel %s disconnected",
                        new_md.name), 0)
                elseif ffi.errno() == C.EAGAIN then
                    -- Retry on this
                else
                    t:stop(l)
                    error(string.format(
                        "Error on connection for channel %s: %s",
                        new_md.name, ffi.string(C.strerror(ffi.errno()))))
                end
            end, sock.sock, ev.WRITE):start(ev.Loop.default)
    end

    -- Function used to write the start time to the proxy
    local function set_start_time(start_time)
        if sock2.sock < 0 then error(string.format(
                "Start time connection for channel is not complete"), 0)
            return
        end
        local msg = string.format("POST /start_time HTTP/1.0\r\n" ..
            "Content-length: %d\r\n\r\n%s", start_time:len(), start_time)

        if start_time ~= 'ABORT' then
            libte.proxyNode_SetStartTime(pn, start_time)

            ev.IO.new(function(l, t, e)
                    local written = C.write(sock2.sock, msg, #msg)
                    if written > 0 then
                        msg = string.sub(msg, written + 1)
                        if #msg == 0 then
                            C.close(sock2.sock)
                            t:stop(l)
                        end
                    elseif written == 0 then
                        t:stop(l)
                        error(string.format(
                            "Start time connection disconnected"), 0)
                    elseif ffi.errno() == C.EAGAIN then
                        -- Retry on this
                    else
                        t:stop(l)
                        error(string.format(
                            "Error on start time connection for channel: %s",
                            ffi.string(C.strerror(ffi.errno()))))
                    end
                end, sock2.sock, ev.WRITE):start(ev.Loop.default)
        else
            -- If we are aborting don't return until the write is done so the
            -- test engine doesn't exit.
            while #msg > 0 do
                local written = C.write(sock2.sock, msg, #msg)
                if written <= 0 then break end
                msg = string.sub(msg, written + 1)
            end
        end
    end

    -- Special for MX878
    if md.connector:find('mx878') then
        start = function(new_md)
            if sock.sock < 0 then
                error(string.format("Connection for channel %s is not complete",
                    new_md.name), 0)
                return
            end
            local out_md = cjson.encode(new_md)
            local msg = string.format("POST /channel_stream HTTP/1.0\r\n" ..
                "Content-length: %d\r\n\r\n%s", #out_md, out_md)

            ev.IO.new(function(l, t, e)
                    local written = C.write(sock.sock, msg, #msg)
                    if written > 0 then
                        msg = string.sub(msg, written + 1)
                        if #msg == 0 then
                            t:stop(l)
                        end
                    elseif written == 0 then
                        t:stop(l)
                        error(string.format("Proxy for channel %s disconnected",
                            new_md.name), 0)
                    elseif ffi.errno() == C.EAGAIN then
                        -- Retry on this
                    else
                        t:stop(l)
                        error(string.format(
                            "Error on connection for channel %s: %s",
                            new_md.name, ffi.string(C.strerror(ffi.errno()))))
                    end
                end, sock.sock, ev.WRITE):start(ev.Loop.default)
            end

    -- Special for MX471
    elseif md.connector:find('mx471') then
        start = function(new_md)
            if sock.sock < 0 then
                error(string.format("Connection for channel %s is not complete",
                    new_md.name), 0)
                return
            end
            local start_time = new_md.start_time
            libte.proxyNode_SetStartTime(pn, nil)

            local cs_node = pn.cs_list
            local md_array = {}
            while cs_node ~= nil do
                local md = cjson.decode(ffi.string(cs_node.stanza))
                md.start_time = start_time
                md_array[#md_array + 1] = md
                cs_node = cs_node.next
            end
            local out_md = cjson.encode(md_array)

            local msg = string.format("POST /channel_stream HTTP/1.0\r\n" ..
                "Content-length: %d\r\n\r\n%s", #out_md, out_md)

            ev.IO.new(function(l, t, e)
                    local written = C.write(sock.sock, msg, #msg)
                    if written > 0 then
                        msg = string.sub(msg, written + 1)
                        if #msg == 0 then
                            t:stop(l)
                        end
                    elseif written == 0 then
                        t:stop(l)
                        error(string.format("Proxy for channel %s disconnected",
                            new_md.name), 0)
                    elseif ffi.errno() == C.EAGAIN then
                        -- Retry on this
                    else
                        t:stop(l)
                        error(string.format(
                            "Error on connection for channel %s: %s",
                            new_md.name, ffi.string(C.strerror(ffi.errno()))))
                    end
                end, sock.sock, ev.WRITE):start(ev.Loop.default)
        end
    end

    -- Establish the TCP connections
    local ok, err = pcall(sock.connect, sock)
    if not ok then
        -- UNIX Domain connect failed, try TCP/IP
        -- Prevents nil to net_ffi on close of socket for network nodes
        -- sock:close()
        sock = net.socket{ip = ip, port = port}
        local ok, err = pcall(sock.connect, sock)
        if not ok then
            sock:close()
            error(string.format("Unable to connect to %s", md.name))
            return
        end
        sock2 = net.socket{ip = ip, port = port}
        local ok, err = pcall(sock2.connect, sock2)
        if not ok then
            sock:close()
            sock2:close()
            error(string.format("Unable to connect to %s", md.name))
            return
        end
    else
        local ok, err = pcall(sock2.connect, sock2)
        if not ok then
            sock:close()
            sock2:close()
            error(string.format("Unable to connect to %s for start time",
                md.name))
            return
        end
    end

    if md.connector:find('mx878') then
        -- All that's needed for an MX878 node is the start function.
        return {state = 2}, start
    end

    pn = libte.proxyNode_Create(cjson.encode(md), sock.sock, pipe_rate)
    if pn == nil then
        sock:close()
        error(string.format("Unable to create proxy node for %s", md.name))
        return
    end

    return pn, start, set_start_time
end

return m
