-- Functions used by the test engine to support automatic SIE file upload

local function signal_uploader(pid)
    -- A bit of delay is required so the uploader process is
    -- ready to accept the signal to make it commence the
    -- upload.
    local d
    ev.Timer.new(function(l, t, e)
            t:stop(l)
            C.kill(ffi.new('pid_t', tonumber(pid)), 10)
            d = true
        end, .5, 0):start(ev.Loop.default)
    while not d do ev.Loop.default:loop('once') end
end

local m = {}

function m:do_auto_upload()

    if not self.sie_auto_upload then return end

    local function get_pref_info()
        local db = sqlite3.open('/var/lib/armada/setup.db')
        if not db then return end
        local si
        local sok = false
        for r in db:rows(string.format(
            [[SELECT content FROM tsd_users WHERE db_name = 'ftp_servers' AND
            deleted = '';]])) do
            si = cjson.decode(r[1])
            if si and si.active and si.username and si.password and
                si.server_ip then
                sok = true
                break
            end
        end
        return sok and si
    end

    self.sie_auto_upload.ftp = get_pref_info()
    if self.sie_auto_upload and not self.sie_auto_upload.ftp then
        -- Auto upload is enabled but there is no "active" FTP server
        log:log(log.NOTICE,
            "FTP upload is enabled but there is no active FTP server")
        self:send_notification_email('event', 'no_active_ftp_server')
        return
    end

    local now = ts.get_ts()

    -- Do automatic SIE upload if configured to do so
    if self.sie_auto_upload and self.sie_auto_upload.ftp and
        self.sie_auto_upload.ftp.username and
        self.sie_auto_upload.ftp.password and
        self.sie_auto_upload.ftp.server_ip then

        -- If the file naming scheme is date/time, use it as-is as the
        -- FTP filename; otherwise (numeric increment filenames), tack
        -- on a timestamp to ensure uniqueness.
        local remote_name
        if self.summary and
                self.summary.sie_filename_mode == 'utc_date_time' then
            remote_name = self.friendly_sie_name .. ".sie"
        else
            remote_name = self.friendly_sie_name .. "" ..
                ts.ts_to_iso(ts.get_ts()) .. ".sie"
        end

        -- FTP doesn't like ":" in the file name so we replace them with "-"
        remote_name = string.gsub(remote_name, ':', '-')

        log:notice(
            '{0} Spawning process to upload SIE {1}.sie file to {2} now = {3}',
            self.test_name, remote_name,
            self.sie_auto_upload.ftp.server_ip,
            ts.epoch_to_iso(now.s + now.ns / 1000000000))
        local email_recipients = self.email_info and self.email_info.to and
            table.concat(self.email_info.to, ',')

        -- If a remote directory path was specified, prepend it to the remote
        -- file name
        if self.sie_auto_upload.ftp.destination_path and
            type(self.sie_auto_upload.ftp.destination_path) == 'string' and
            #self.sie_auto_upload.ftp.destination_path > 0 then
            if not self.sie_auto_upload.ftp.destination_path:find("/$") then
                self.sie_auto_upload.ftp.destination_path =
                    self.sie_auto_upload.ftp.destination_path .. '/'
            end
            remote_name = self.sie_auto_upload.ftp.destination_path ..
                remote_name
        end

        local pid = master:SpawnNonCritical(
            '/usr/bin/upload-sie',
            {
                '/usr/bin/upload-sie',
                remote_name,
                self.sie_name,
                self.sie_auto_upload.ftp.post_remove and 'remove' or 'keep',
                'ftp',
                self.sie_auto_upload.ftp.username,
                self.sie_auto_upload.ftp.password,
                self.sie_auto_upload.ftp.server_ip,
                email_recipients,
            })
        if not pid then
            error(string.format(
                "Unable to upload SIE file %s to %s: " ..
                "Unable to spawn upload process",
                self.friendly_sie_name,
                self.sie_auto_upload.ftp.server_ip), true)
        elseif not self.sie_auto_upload.delay then
            -- Signal the uploader to upload the SIE file immediately
            signal_uploader(pid)
        else
            -- If automatic upload is enabled AND it is configured
            -- to upload after a delay when the next cycle starts
            -- the signal sent to the upload process will be sent
            -- after the desired delay when that cycle starts.
        end
    end
end

return m.do_auto_upload
